#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from xpra.client.mixins.notification import NotificationClient
from unit.client.mixins.clientmixintest_util import ClientMixinTest


class NotificationClientTest(ClientMixinTest):

	def test_notification(self):
		opts = AdHocStruct()
		opts.notifications = True
		self._test_mixin_class(NotificationClient, opts, {
			"notifications" : True,
			"notifications.close" : True,
			})

def main():
	unittest.main()


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from enum import IntEnum
from collections.abc import Callable

from xpra.dbus.common import loop_init, init_session_bus
from xpra.log import Logger

log = Logger("shadow")

BASE_REQUEST_PATH = "/org/freedesktop/portal/desktop/request"
PORTAL_REQUEST = "org.freedesktop.portal.Request"
PORTAL_SESSION_INTERFACE = "org.freedesktop.portal.Session"
PORTAL_DESKTOP_INTERFACE = "org.freedesktop.portal.Desktop"
PORTAL_DESKTOP_PATH = "/org/freedesktop/portal/desktop"
REMOTEDESKTOP_IFACE = "org.freedesktop.portal.RemoteDesktop"
SCREENCAST_IFACE = "org.freedesktop.portal.ScreenCast"

loop_init()
bus = init_session_bus()


class AvailableDeviceTypes(IntEnum):
    KEYBOARD = 1
    POINTER = 2
    TOUCHSCREEN = 4


class AvailableSourceTypes(IntEnum):
    MONITOR = 1
    WINDOW = 2
    VIRTUAL = 4


dbus_sender_name: str = (bus.get_unique_name()[1:]).replace(".", "_")
request_counter: int = 0


def screenscast_dbus_call(method, callback: Callable, *args, options=None) -> None:
    dbus_desktop_call(SCREENCAST_IFACE, method, callback, *args, options=options)


def remotedesktop_dbus_call(method, callback: Callable, *args, options=None) -> None:
    dbus_desktop_call(REMOTEDESKTOP_IFACE, method, callback, *args, options=options)


def dbus_desktop_call(interface: str, method, callback: Callable, *args, options=None) -> None:
    # generate a new token and path:
    options = options or {}
    global request_counter
    request_counter += 1
    request_token = f"u{request_counter}"
    request_path = f"{BASE_REQUEST_PATH}/{dbus_sender_name}/{request_token}"
    log(f"adding dbus signal receiver {callback}")
    bus.add_signal_receiver(callback,
                            "Response",
                            PORTAL_REQUEST,
                            PORTAL_DESKTOP_INTERFACE,
                            request_path)
    options["handle_token"] = request_token
    log(f"calling {method} with args={args}, options={options}")
    method(*(args + (options,)), dbus_interface=interface)


def get_portal_interface():
    return bus.get_object(PORTAL_DESKTOP_INTERFACE, PORTAL_DESKTOP_PATH)


def get_session_interface(session_path: str):
    return bus.get_object(PORTAL_DESKTOP_INTERFACE, session_path)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from ctypes import (
    WinDLL, WINFUNCTYPE, GetLastError,  # @UnresolvedImport
    POINTER, Structure,
    c_ulong, c_ushort, c_ubyte, c_int, c_long, c_void_p, c_size_t, c_char,
    byref, sizeof,
)
from ctypes.wintypes import (
    HWND, DWORD, WPARAM, LPARAM, HDC, HMONITOR, HMODULE,
    SHORT, ATOM, RECT, POINT, MAX_PATH, WCHAR, BYTE,
    HANDLE, LPSTR, LPCWSTR, UINT, INT, BOOL, WORD, HGDIOBJ,
    LONG, LPVOID, HBITMAP, LPCSTR, LPWSTR, HWINSTA,
    HINSTANCE, HMENU, ULONG, HHOOK, LPMSG,
    LCID, HKL,
)

# imported from this module but not used here:
assert GetLastError

LPCTSTR = LPCSTR
LRESULT = c_long
DEVMODE = c_void_p
PDWORD = POINTER(DWORD)
LPDWORD = POINTER(DWORD)
PBYTE = POINTER(BYTE)
HCURSOR = HANDLE
HICON = HANDLE
HBRUSH = HANDLE


class CURSORINFO(Structure):
    _fields_ = [
        ("cbSize", DWORD),
        ("flags", DWORD),
        ("hCursor", HCURSOR),
        ("ptScreenPos", POINT),
    ]


class ICONINFO(Structure):
    _fields_ = [
        ("fIcon", BOOL),
        ("xHotspot", DWORD),
        ("yHotspot", DWORD),
        ("hbmMask", HBITMAP),
        ("hbmColor", HBITMAP),
    ]


PICONINFO = POINTER(ICONINFO)


# noinspection PyTypeChecker
class ICONINFOEXA(Structure):
    _fields_ = [
        ("cbSize", DWORD),
        ("fIcon", BOOL),
        ("xHotspot", DWORD),
        ("yHotspot", DWORD),
        ("hbmMask", HBITMAP),
        ("hbmColor", HBITMAP),
        ("wResID", WORD),
        ("sxModName", c_char * MAX_PATH),
        ("szResName", c_char * MAX_PATH),
    ]


PICONINFOEXA = POINTER(ICONINFOEXA)


# noinspection PyTypeChecker
class ICONINFOEXW(Structure):
    _fields_ = [
        ("cbSize", DWORD),
        ("fIcon", BOOL),
        ("xHotspot", DWORD),
        ("yHotspot", DWORD),
        ("hbmMask", HBITMAP),
        ("hbmColor", HBITMAP),
        ("wResID", WORD),
        ("sxModName", WCHAR * MAX_PATH),
        ("szResName", WCHAR * MAX_PATH),
    ]


PICONINFOEXW = POINTER(ICONINFOEXW)


class Bitmap(Structure):
    _fields_ = [
        ("bmType", LONG),
        ("bmWidth", LONG),
        ("bmHeight", LONG),
        ("bmWidthBytes", LONG),
        ("bmPlanes", WORD),
        ("bmBitsPixel", WORD),
        ("bmBits", LPVOID)
    ]


class CIEXYZ(Structure):
    _fields_ = [
        ('ciexyzX', DWORD),
        ('ciexyzY', DWORD),
        ('ciexyzZ', DWORD),
    ]


class CIEXYZTRIPLE(Structure):
    _fields_ = [
        ('ciexyzRed', CIEXYZ),
        ('ciexyzBlue', CIEXYZ),
        ('ciexyzGreen', CIEXYZ),
    ]


class BITMAPINFOHEADER(Structure):
    _fields_ = [
        ("biSize", DWORD),
        ("biWidth", LONG),
        ("biHeight", LONG),
        ("biPlanes", WORD),
        ("biBitCount", WORD),
        ("biCompression", DWORD),
        ("biSizeImage", DWORD),
        ("biXPelsPerMeter", LONG),
        ("biYPelsPerMeter", LONG),
        ("biClrUsed", DWORD),
        ("biClrImportant", DWORD),
    ]


PBITMAPINFOHEADER = POINTER(BITMAPINFOHEADER)


class BITMAPINFO(Structure):
    _fields_ = [
        ("bmiHeader", BITMAPINFOHEADER),
        ("bmiColors", DWORD),
    ]


PBITMAPINFO = POINTER(BITMAPINFO)


class BITMAPV5HEADER(Structure):
    _fields_ = [
        ('bV5Size', DWORD),
        ('bV5Width', LONG),
        ('bV5Height', LONG),
        ('bV5Planes', WORD),
        ('bV5BitCount', WORD),
        ('bV5Compression', DWORD),
        ('bV5SizeImage', DWORD),
        ('bV5XPelsPerMeter', LONG),
        ('bV5YPelsPerMeter', LONG),
        ('bV5ClrUsed', DWORD),
        ('bV5ClrImportant', DWORD),
        ('bV5RedMask', DWORD),
        ('bV5GreenMask', DWORD),
        ('bV5BlueMask', DWORD),
        ('bV5AlphaMask', DWORD),
        ('bV5CSType', DWORD),
        ('bV5Endpoints', CIEXYZTRIPLE),
        ('bV5GammaRed', DWORD),
        ('bV5GammaGreen', DWORD),
        ('bV5GammaBlue', DWORD),
        ('bV5Intent', DWORD),
        ('bV5ProfileData', DWORD),
        ('bV5ProfileSize', DWORD),
        ('bV5Reserved', DWORD),
    ]


PBITMAPV5HEADER = POINTER(BITMAPV5HEADER)

CCHDEVICENAME = 32


class MONITORINFOEX(Structure):
    # noinspection PyTypeChecker
    _fields_ = [
        ('cbSize', DWORD),
        ('rcMonitor', RECT),
        ('rcWork', RECT),
        ('dwFlags', DWORD),
        ('szDevice', WCHAR * CCHDEVICENAME),
    ]


class SECURITY_ATTRIBUTES(Structure):
    _fields_ = [
        ("nLength", c_int),
        ("lpSecurityDescriptor", c_void_p),
        ("bInheritHandle", c_int),
    ]


LPSECURITY_ATTRIBUTES = POINTER(SECURITY_ATTRIBUTES)


def GetMonitorInfo(hmonitor) -> dict[str, Any]:
    info = MONITORINFOEX()
    info.szDevice = ""
    info.cbSize = sizeof(MONITORINFOEX)
    if not GetMonitorInfoW(hmonitor, byref(info)):
        raise OSError()  # @UndefinedVariable
    monitor = info.rcMonitor.left, info.rcMonitor.top, info.rcMonitor.right, info.rcMonitor.bottom
    work = info.rcWork.left, info.rcWork.top, info.rcWork.right, info.rcWork.bottom
    return {
        "Work": work,
        "Monitor": monitor,
        "Flags": info.dwFlags,
        "Device": info.szDevice or "",
    }


kernel32 = WinDLL("kernel32", use_last_error=True)
LocalFree = kernel32.LocalFree
FormatMessageW = kernel32.FormatMessageW
GetCurrentThreadId = kernel32.GetCurrentThreadId
GetCurrentThreadId.restype = DWORD
GetCurrentThreadId.argtypes = []
SetConsoleTitleA = kernel32.SetConsoleTitleA
SetConsoleTitleA.restype = INT
SetConsoleTitleA.argtypes = [LPCTSTR]
GetConsoleScreenBufferInfo = kernel32.GetConsoleScreenBufferInfo
GetModuleHandleA = kernel32.GetModuleHandleA
GetModuleHandleA.restype = HMODULE
GetModuleHandleW = kernel32.GetModuleHandleW
GetModuleHandleW.restype = HMODULE
GetModuleHandleExA = kernel32.GetModuleHandleExA
GetModuleHandleExA.argtypes = [DWORD, LPCSTR, POINTER(HMODULE)]
GetModuleHandleExA.restype = BOOL
ConsoleCtrlHandler = WINFUNCTYPE(BOOL, DWORD)
SetConsoleCtrlHandler = kernel32.SetConsoleCtrlHandler
SetConsoleCtrlHandler.argtypes = (ConsoleCtrlHandler, BOOL)
SetConsoleCtrlHandler.restype = BOOL
GetComputerNameW = kernel32.GetComputerNameW
GetComputerNameW.restype = BOOL
GetComputerNameW.argtypes = [LPWSTR, LPDWORD]
GetCurrentProcess = kernel32.GetCurrentProcess
GetCurrentProcess.argtypes = []
GetCurrentProcess.restype = HANDLE
HeapAlloc = kernel32.HeapAlloc
HeapAlloc.restype = LPVOID
HeapAlloc.argtypes = [HANDLE, DWORD, c_size_t]
GetProcessHeap = kernel32.GetProcessHeap
GetProcessHeap.restype = HANDLE
GetProcessHeap.argtypes = []
CloseHandle = kernel32.CloseHandle
CloseHandle.argtypes = [HANDLE]
CloseHandle.restype = BOOL
GetProductInfo = kernel32.GetProductInfo
GetProductInfo.argtypes = [DWORD, DWORD, DWORD, DWORD, PDWORD]
GetProductInfo.restype = BOOL
GetStdHandle = WINFUNCTYPE(HANDLE, DWORD)(("GetStdHandle", kernel32))
HGLOBAL = HANDLE
GlobalAlloc = kernel32.GlobalAlloc
GlobalAlloc.restype = HGLOBAL
GlobalAlloc.argtypes = [UINT, c_size_t]
GlobalFree = kernel32.GlobalFree
GlobalFree.restype = HGLOBAL
GlobalFree.argtypes = [HGLOBAL]
GlobalLock = kernel32.GlobalLock
GlobalLock.restype = LPVOID
GlobalLock.argtypes = [HGLOBAL]
GlobalSize = kernel32.GlobalSize
GlobalSize.argtypes = [HGLOBAL]
GlobalSize.restype = c_size_t
GlobalUnlock = kernel32.GlobalUnlock
GlobalUnlock.restype = BOOL
GlobalUnlock.argtypes = [HGLOBAL]
WideCharToMultiByte = kernel32.WideCharToMultiByte
WideCharToMultiByte.restype = c_int
WideCharToMultiByte.argtypes = [UINT, DWORD, LPCWSTR, c_int, c_void_p, c_int, LPCSTR, POINTER(BOOL)]
MultiByteToWideChar = kernel32.MultiByteToWideChar
MultiByteToWideChar.restype = c_int
MultiByteToWideChar.argtypes = [UINT, DWORD, LPCSTR, c_int, LPWSTR, c_int]
WTSGetActiveConsoleSessionId = kernel32.WTSGetActiveConsoleSessionId
WTSGetActiveConsoleSessionId.restype = DWORD
WTSGetActiveConsoleSessionId.argtypes = []
QueryFullProcessImageNameA = kernel32.QueryFullProcessImageNameA
QueryFullProcessImageNameA.restype = BOOL
QueryFullProcessImageNameA.argtypes = [HANDLE, DWORD, LPSTR, PDWORD]
OpenProcess = kernel32.OpenProcess
OpenProcess.restype = HANDLE
OpenProcess.argtypes = [DWORD, BOOL, DWORD]
GetThreadLocale = kernel32.GetThreadLocale
GetThreadLocale.argtypes = []
GetThreadLocale.restype = LCID

user32 = WinDLL("user32", use_last_error=True)
RegisterClassExA = user32.RegisterClassExA
RegisterClassExA.restype = ATOM
RegisterClassExW = user32.RegisterClassExW
RegisterClassExW.restype = ATOM
UnregisterClassA = user32.UnregisterClassA
UnregisterClassA.restype = BOOL
UnregisterClassA.argtypes = [LPCSTR, HINSTANCE]
UnregisterClassW = user32.UnregisterClassW
UnregisterClassW.restype = BOOL
UnregisterClassW.argtypes = [LPCWSTR, HINSTANCE]
CreateWindowExA = user32.CreateWindowExA
CreateWindowExA.restype = HWND
CreateWindowExA.argtypes = [DWORD, ATOM, LPCWSTR, DWORD, c_int, c_int, c_int, c_int, HWND, HMENU, HINSTANCE, LPVOID]
CreateWindowExW = user32.CreateWindowExW
CreateWindowExW.restype = HWND
CreateWindowExW.argtypes = [DWORD, ATOM, LPCWSTR, DWORD, c_int, c_int, c_int, c_int, HWND, HMENU, HINSTANCE, LPVOID]
DestroyWindow = user32.DestroyWindow
DestroyWindow.restype = BOOL
DestroyWindow.argtypes = [HWND]
DefWindowProcA = user32.DefWindowProcA
DefWindowProcA.argtypes = [HWND, UINT, WPARAM, LPARAM]
DefWindowProcA.restype = INT
DefWindowProcW = user32.DefWindowProcW
DefWindowProcW.argtypes = [HWND, UINT, WPARAM, LPARAM]
DefWindowProcW.restype = INT
WNDPROC = WINFUNCTYPE(LRESULT, HWND, UINT, WPARAM, LPARAM)
CallWindowProcW = user32.CallWindowProcW
CallWindowProcW.argtypes = [WNDPROC, HWND, UINT, WPARAM, LPARAM]
CallWindowProcW.restype = LRESULT
MessageBoxA = user32.MessageBoxA
MessageBoxA.restype = INT
MessageBoxA.argtypes = [HWND, LPCTSTR, LPCTSTR, UINT]
GetSystemMetrics = user32.GetSystemMetrics
GetSystemMetrics.restype = INT
GetSystemMetrics.argtypes = [INT]
SetWindowLongA = user32.SetWindowLongA
SetWindowLongA.restype = LONG
SetWindowLongA.argtypes = [HWND, INT, LONG]
SetWindowLongW = user32.SetWindowLongW
SetWindowLongW.restype = LONG
SetWindowLongW.argtypes = [HWND, INT, WNDPROC]
GetWindowLongW = user32.GetWindowLongW
GetWindowLongW.restype = LONG
GetWindowLongW.argtypes = [HWND, INT]
ClipCursor = user32.ClipCursor
ClipCursor.restype = BOOL
GetCursorPos = user32.GetCursorPos
GetCursorPos.argtypes = [POINTER(POINT)]
GetCursorPos.restype = BOOL
SetCursorPos = user32.SetCursorPos
SetCursorPos.argtypes = [INT, INT]
SetCursorPos.restype = BOOL
GetPhysicalCursorPos = user32.GetPhysicalCursorPos
GetPhysicalCursorPos.argtypes = [POINTER(POINT)]
GetPhysicalCursorPos.restype = BOOL
SetPhysicalCursorPos = user32.SetPhysicalCursorPos
SetPhysicalCursorPos.argtypes = [INT, INT]
SetPhysicalCursorPos.restype = BOOL
GetCursorInfo = user32.GetCursorInfo
GetCursorInfo.argtypes = [POINTER(CURSORINFO)]
GetCursorInfo.restype = BOOL
LogicalToPhysicalPoint = user32.LogicalToPhysicalPoint
LogicalToPhysicalPoint.argtypes = [HWND, POINTER(POINT)]
LogicalToPhysicalPoint.restype = BOOL
SendMessageA = user32.SendMessageA
PostMessageA = user32.PostMessageA
FindWindowA = user32.FindWindowA
GetWindowRect = user32.GetWindowRect
GetDoubleClickTime = user32.GetDoubleClickTime
_EnumDisplayMonitors = user32.EnumDisplayMonitors
MonitorFromWindow = user32.MonitorFromWindow
MonitorFromWindow.restype = HMONITOR
GetMonitorInfoW = user32.GetMonitorInfoW
GetMonitorInfoW.argtypes = [HMONITOR, POINTER(MONITORINFOEX)]
GetMonitorInfoW.restype = BOOL
UnhookWindowsHookEx = user32.UnhookWindowsHookEx
CallNextHookEx = user32.CallNextHookEx
CallNextHookEx.restype = LRESULT
CallNextHookEx.argtyps = [HHOOK, INT, WPARAM, LPARAM]
SetWindowsHookExA = user32.SetWindowsHookExA
SetWindowsHookExA.restype = HHOOK
SetWindowsHookExA.argtypes = [INT, HANDLE, HINSTANCE, DWORD]
GetMessageA = user32.GetMessageA
GetMessageA.restype = BOOL
GetMessageA.argtypes = [LPMSG, HWND, UINT, UINT]
TranslateMessage = user32.TranslateMessage
TranslateMessage.restype = BOOL
TranslateMessage.argtypes = [LPMSG]
DispatchMessageA = user32.DispatchMessageA
DispatchMessageA.restype = LRESULT
DispatchMessageA.argtypes = [LPMSG]
MapVirtualKeyW = user32.MapVirtualKeyW
GetKeyboardState = user32.GetKeyboardState
GetKeyboardState.argtypes = [PBYTE]
GetKeyboardState.restype = BOOL
SetKeyboardState = user32.SetKeyboardState
SetKeyboardState.argtypes = [PBYTE]
SetKeyboardState.restype = BOOL
GetAsyncKeyState = user32.GetAsyncKeyState
GetAsyncKeyState.argtypes = [INT]
GetAsyncKeyState.restype = SHORT
VkKeyScanW = user32.VkKeyScanW
VkKeyScanW.argtypes = [WCHAR]
keybd_event = user32.keybd_event
GetKeyState = user32.GetKeyState
GetKeyState.restype = SHORT
GetKeyboardLayout = user32.GetKeyboardLayout
GetKeyboardLayout.argtypes = [DWORD]
GetKeyboardLayout.restype = HKL
GetKeyboardLayoutList = user32.GetKeyboardLayoutList
GetKeyboardLayoutList.restype = int
# noinspection PyTypeChecker
GetKeyboardLayoutList.argtypes = [c_int, POINTER(HANDLE * 32)]
GetKeyboardLayoutName = user32.GetKeyboardLayoutNameA
GetKeyboardLayoutName.restype = BOOL
GetKeyboardLayoutName.argtypes = [LPSTR]
ActivateKeyboardLayout = user32.ActivateKeyboardLayout
ActivateKeyboardLayout.argtypes = [HKL, UINT]
ActivateKeyboardLayout.restype = HKL
SystemParametersInfoA = user32.SystemParametersInfoA
EnumWindowsProc = WINFUNCTYPE(BOOL, HWND, LPARAM)
EnumWindows = user32.EnumWindows
EnumWindows.argtypes = [EnumWindowsProc, LPARAM]
IsWindowVisible = user32.IsWindowVisible
GetWindowThreadProcessId = user32.GetWindowThreadProcessId
GetWindowThreadProcessId.argtypes = [HWND, LPDWORD]
GetWindowThreadProcessId.restype = DWORD
GetDesktopWindow = user32.GetDesktopWindow
GetDesktopWindow.restype = HWND
GetWindowDC = user32.GetWindowDC
GetWindowDC.argtypes = [HWND]
GetWindowDC.restype = HDC
ReleaseDC = user32.ReleaseDC
ReleaseDC.restype = c_int
ReleaseDC.argtypes = [HWND, HDC]
mouse_event = user32.mouse_event
LoadIconA = user32.LoadIconA
LoadIconA.restype = HICON
# can also pass int as second arg, so don't declare argtypes:
# LoadIconA.argtypes = [HINSTANCE, LPCSTR]
RegisterWindowMessageA = user32.RegisterWindowMessageA
RegisterWindowMessageA.restype = UINT
RegisterWindowMessageA.argtypes = [LPCSTR]
UpdateWindow = user32.UpdateWindow
UpdateWindow.restype = BOOL
UpdateWindow.argtypes = [HWND]
DestroyIcon = user32.DestroyIcon
DestroyIcon.restype = BOOL
DestroyIcon.argtypes = [HICON]
LoadImageW = user32.LoadImageW
LoadImageW.argtypes = [HINSTANCE, LPCWSTR, UINT, INT, INT, UINT]
LoadImageW.restype = HANDLE
CreateIconIndirect = user32.CreateIconIndirect
CreateIconIndirect.restype = HICON
CreateIconIndirect.argtypes = [POINTER(ICONINFO)]
GetDC = user32.GetDC
GetDC.argtypes = [HWND]
GetDC.restype = HDC
ReleaseDC = user32.ReleaseDC
ReleaseDC.argtypes = [HWND, HDC]
ReleaseDC.restype = c_int
DrawIcon = user32.DrawIcon
DrawIcon.argtypes = [HDC, INT, INT, HICON]
DrawIcon.restype = BOOL
DrawIconEx = user32.DrawIconEx
DrawIconEx.argtypes = [HDC, INT, INT, HICON, INT, INT, UINT, HBRUSH, UINT]
DrawIconEx.restype = BOOL
GetIconInfo = user32.GetIconInfo
GetIconInfo.argtypes = [HICON, PICONINFO]
GetIconInfo.restype = BOOL
GetIconInfoExA = user32.GetIconInfoExA
GetIconInfoExA.argtypes = [HICON, PICONINFOEXA]
GetIconInfoExA.restype = BOOL
GetIconInfoExW = user32.GetIconInfoExW
GetIconInfoExW.argtypes = [HICON, PICONINFOEXW]
GetIconInfoExW.restype = BOOL
CopyIcon = user32.CopyIcon
CopyIcon.argtypes = [HICON]
CopyIcon.restype = HICON
PostQuitMessage = user32.PostQuitMessage
PostQuitMessage.argtypes = [INT]
OpenWindowStationW = user32.OpenWindowStationW
OpenWindowStationW.restype = HWINSTA
ACCESS_MASK = DWORD
OpenWindowStationW.argtypes = [LPWSTR, BOOL, ACCESS_MASK]
OpenWindowStationA = user32.OpenWindowStationA
OpenWindowStationA.restype = HWINSTA
OpenWindowStationA.argtypes = [LPCSTR, BOOL, ACCESS_MASK]
GetProcessWindowStation = user32.GetProcessWindowStation
GetProcessWindowStation.restype = HWINSTA
GetProcessWindowStation.argtypes = []
SetProcessWindowStation = user32.SetProcessWindowStation
SetProcessWindowStation.restype = BOOL
SetProcessWindowStation.argtypes = [HWINSTA]
CreateWindowStationA = user32.CreateWindowStationA
CreateWindowStationA.restype = HWINSTA
CreateWindowStationA.argtypes = [LPCSTR, DWORD, ACCESS_MASK, LPSECURITY_ATTRIBUTES]
CloseWindowStation = user32.CloseWindowStation
CloseWindowStation.restype = BOOL
CloseWindowStation.argtypes = [HWINSTA]
WINSTAENUMPROCA = WINFUNCTYPE(BOOL, LPSTR, LPARAM)
EnumWindowStationsA = user32.EnumWindowStationsA
EnumWindowStationsA.restype = BOOL
EnumWindowStationsA.argtypes = [WINSTAENUMPROCA, LPARAM]
HDESK = HANDLE
CreateDesktopA = user32.CreateDesktopA
CreateDesktopA.restype = HDESK
DEVMODEA = c_void_p
CreateDesktopA.argtypes = [LPCSTR, LPCSTR, DEVMODEA, DWORD, ACCESS_MASK, LPSECURITY_ATTRIBUTES]
CreateDesktopExA = user32.CreateDesktopExA
CreateDesktopExA.argtypes = [LPCSTR, LPCSTR, DEVMODEA, DWORD, ACCESS_MASK, LPSECURITY_ATTRIBUTES, ULONG, LPVOID]
CreateDesktopExA = user32.CreateDesktopExA
CreateDesktopExA.restype = HDESK
OpenDesktopW = user32.OpenDesktopW
OpenDesktopW.restype = HDESK
OpenDesktopW.argtypes = [LPWSTR, DWORD, BOOL, ACCESS_MASK]
CloseDesktop = user32.CloseDesktop
CloseDesktop.restype = BOOL
CloseDesktop.argtypes = [HDESK]
OpenInputDesktop = user32.OpenInputDesktop
OpenInputDesktop.restype = HDESK
OpenInputDesktop.argtypes = [DWORD, BOOL, ACCESS_MASK]
DESKTOPENUMPROCA = WINFUNCTYPE(BOOL, LPCSTR, LPARAM)
EnumDesktopsA = user32.EnumDesktopsA
EnumDesktopsA.restype = BOOL
EnumDesktopsA.argtypes = [HWINSTA, DESKTOPENUMPROCA, LPARAM]
GetThreadDesktop = user32.GetThreadDesktop
GetThreadDesktop.restype = HDESK
GetThreadDesktop.argtypes = [DWORD]
GetUserObjectInformationA = user32.GetUserObjectInformationA
GetUserObjectInformationA.restype = BOOL
GetUserObjectInformationA.argtypes = [HANDLE, INT, LPVOID, DWORD, LPDWORD]
CreatePopupMenu = user32.CreatePopupMenu
CreatePopupMenu.restype = HMENU
CreatePopupMenu.argtypes = []
AppendMenu = user32.AppendMenuW
AppendMenu.restype = BOOL
AppendMenu.argtypes = [HMENU, UINT, UINT, LPCWSTR]


# clipboard functions:


class COPYDATASTRUCT(Structure):
    _fields_ = [
        ("dwData", POINTER(ULONG)),
        ("cbData", DWORD),
        ("lpData", LPVOID),
    ]


OpenClipboard = user32.OpenClipboard
OpenClipboard.restype = BOOL
OpenClipboard.argtypes = [HWND]
CloseClipboard = user32.CloseClipboard
CloseClipboard.restype = BOOL
CloseClipboard.argtypes = []
EmptyClipboard = user32.EmptyClipboard
EmptyClipboard.restype = BOOL
EmptyClipboard.argtypes = []
SetClipboardData = user32.SetClipboardData
SetClipboardData.restype = HANDLE
SetClipboardData.argtypes = [UINT, HANDLE]
RegisterClipboardFormatA = user32.RegisterClipboardFormatA
RegisterClipboardFormatA.restype = UINT
RegisterClipboardFormatA.argtypes = [LPCSTR]
GetClipboardData = user32.GetClipboardData
GetClipboardData.restype = HANDLE
GetClipboardData.argtypes = [UINT]
GetClipboardSequenceNumber = user32.GetClipboardSequenceNumber
GetClipboardSequenceNumber.restype = DWORD
GetClipboardSequenceNumber.argtypes = []
GetClipboardOwner = user32.GetClipboardOwner
GetClipboardOwner.restype = HWND
GetClipboardOwner.argtypes = []
GetOpenClipboardWindow = user32.GetOpenClipboardWindow
GetOpenClipboardWindow.restype = HWND
GetOpenClipboardWindow.argtypes = []
AddClipboardFormatListener = user32.AddClipboardFormatListener
AddClipboardFormatListener.restype = BOOL
AddClipboardFormatListener.argtypes = [HWND]
RemoveClipboardFormatListener = user32.RemoveClipboardFormatListener
RemoveClipboardFormatListener.restype = BOOL
RemoveClipboardFormatListener.argtypes = [HWND]
GetUpdatedClipboardFormats = user32.GetUpdatedClipboardFormats
GetUpdatedClipboardFormats.restype = BOOL
GetUpdatedClipboardFormats.argtypes = [POINTER(UINT), UINT, POINTER(UINT)]
IsClipboardFormatAvailable = user32.IsClipboardFormatAvailable
IsClipboardFormatAvailable.restype = BOOL
IsClipboardFormatAvailable.argtypes = [UINT]
SetClipboardViewer = user32.SetClipboardViewer
SetClipboardViewer.restype = HWND
SetClipboardViewer.argtypes = [HWND]
GetClipboardFormatNameA = user32.GetClipboardFormatNameA
GetClipboardFormatNameA.restype = c_int
GetClipboardFormatNameA.argtypes = [UINT, LPCSTR, c_int]
GetClipboardFormatNameW = user32.GetClipboardFormatNameW
GetClipboardFormatNameW.restype = c_int
GetClipboardFormatNameW.argtypes = [UINT, LPWSTR, c_int]
EnumClipboardFormats = user32.EnumClipboardFormats
EnumClipboardFormats.restype = UINT
EnumClipboardFormats.argtypes = [UINT]
CountClipboardFormats = user32.CountClipboardFormats
CountClipboardFormats.restype = c_int
CountClipboardFormats.argtypes = []
GetWindowTextA = user32.GetWindowTextA
GetWindowTextA.restype = c_int
GetWindowTextA.argtypes = [HWND, LPSTR, c_int]
GetWindowTextW = user32.GetWindowTextW
GetWindowTextW.restype = c_int
GetWindowTextW.argtypes = [HWND, LPWSTR, c_int]
GetWindowTextLengthA = user32.GetWindowTextLengthA
GetWindowTextLengthA.restype = c_int
GetWindowTextLengthA.argtypes = [HWND]
GetWindowTextLengthW = user32.GetWindowTextLengthW
GetWindowTextLengthW.restype = c_int
GetWindowTextLengthW.argtypes = [HWND]


class PROCESS_INFORMATION(Structure):
    _fields_ = (
        ('_hProcess', HANDLE),
        ('_hThread', HANDLE),
        ('dwProcessId', DWORD),
        ('dwThreadId', DWORD),
    )


PPROCESS_INFORMATION = POINTER(PROCESS_INFORMATION)


class STARTUPINFOA(Structure):
    _fields_ = (
        ('cb', DWORD),
        ('lpReserved', LPWSTR),
        ('lpDesktop', LPWSTR),
        ('lpTitle', LPWSTR),
        ('dwX', DWORD),
        ('dwY', DWORD),
        ('dwXSize', DWORD),
        ('dwYSize', DWORD),
        ('dwXCountChars', DWORD),
        ('dwYCountChars', DWORD),
        ('dwFillAttribute', DWORD),
        ('dwFlags', DWORD),
        ('wShowWindow', WORD),
        ('cbReserved2', WORD),
        ('lpReserved2', c_void_p),
        ('hStdInput', HANDLE),
        ('hStdOutput', HANDLE),
        ('hStdError', HANDLE),
    )


PSTARTUPINFOA = POINTER(STARTUPINFOA)
advapi32 = WinDLL("advapi32")
CreateProcessAsUserA = advapi32.CreateProcessAsUserA
CreateProcessAsUserA.restype = BOOL
CreateProcessAsUserA.argtypes = [HANDLE, LPCSTR, LPSTR, LPSECURITY_ATTRIBUTES, LPSECURITY_ATTRIBUTES, BOOL, DWORD,
                                 LPVOID, LPCSTR, LPCSTR, PSTARTUPINFOA, PPROCESS_INFORMATION]
LogonUser = advapi32.LogonUserW
LogonUser.argtypes = [LPCWSTR, LPCWSTR, LPCWSTR, DWORD, DWORD, POINTER(HANDLE)]
LogonUser.restype = BOOL
GetUserNameA = advapi32.GetUserNameA
GetUserNameA.restype = BOOL

gdi32 = WinDLL("gdi32", use_last_error=True)
CreateCompatibleDC = gdi32.CreateCompatibleDC
CreateCompatibleDC.argtypes = [HDC]
CreateCompatibleDC.restype = HDC
CreateCompatibleBitmap = gdi32.CreateCompatibleBitmap
CreateCompatibleBitmap.argtypes = [HDC, c_int, c_int]
CreateCompatibleBitmap.restype = HBITMAP
CreateBitmap = gdi32.CreateBitmap
CreateBitmap.argtypes = [INT, INT, UINT, UINT, POINTER(c_void_p)]
CreateBitmap.restype = HBITMAP
GetBitmapBits = gdi32.GetBitmapBits
GetBitmapBits.argtypes = [HGDIOBJ, LONG, LPVOID]
GetBitmapBits.restype = LONG
SelectObject = gdi32.SelectObject
SelectObject.argtypes = [HDC, HGDIOBJ]
SelectObject.restype = HGDIOBJ
BitBlt = gdi32.BitBlt
BitBlt.argtypes = [HDC, c_int, c_int, c_int, c_int, HDC, c_int, c_int, DWORD]
BitBlt.restype = BOOL
GetDeviceCaps = gdi32.GetDeviceCaps
GetDeviceCaps.argtypes = [HDC, c_int]
GetDeviceCaps.restype = c_int
GetSystemPaletteEntries = gdi32.GetSystemPaletteEntries
GetSystemPaletteEntries.restype = UINT
GetStockObject = gdi32.GetStockObject
GetStockObject.restype = HGDIOBJ
SetPixelV = gdi32.SetPixelV
CreateDIBSection = gdi32.CreateDIBSection
CreateDIBSection.restype = HBITMAP
CreateDIBSection.argtypes = [HANDLE, POINTER(BITMAPV5HEADER), UINT, POINTER(c_void_p), HANDLE, DWORD]
CreateDIBitmap = gdi32.CreateDIBitmap
CreateDIBitmap.restype = HBITMAP
CreateDIBitmap.argtypes = [HDC, PBITMAPINFOHEADER, DWORD, c_void_p, PBITMAPINFO, UINT]
DeleteObject = gdi32.DeleteObject
DeleteObject.argtypes = [HGDIOBJ]
DeleteObject.restype = BOOL
DeleteDC = gdi32.DeleteDC
DeleteDC.restype = BOOL
DeleteDC.argtypes = [HDC]
CreateDCA = gdi32.CreateDCA
CreateDCA.restype = HDC
CreateDCA.argtypes = [LPCSTR, LPCSTR, LPCSTR, DEVMODE]
ChoosePixelFormat = gdi32.ChoosePixelFormat
ChoosePixelFormat.argtypes = [HDC, c_void_p]
ChoosePixelFormat.restype = c_int
GetPixelFormat = gdi32.GetPixelFormat
GetPixelFormat.argtypes = [HDC]
GetPixelFormat.restype = c_int
SetPixelFormat = gdi32.SetPixelFormat
SetPixelFormat.argtypes = [HDC, c_int, c_void_p]
SetPixelFormat.restype = BOOL
DescribePixelFormat = gdi32.DescribePixelFormat
DescribePixelFormat.argtypes = [HDC, c_int, UINT, c_void_p]
DescribePixelFormat.restype = c_int
SwapBuffers = gdi32.SwapBuffers
SwapBuffers.argtypes = [HDC]
SwapBuffers.restype = BOOL
BeginPaint = user32.BeginPaint
BeginPaint.argtypes = [HWND, c_void_p]
BeginPaint.restype = HDC
EndPaint = user32.EndPaint
EndPaint.argtypes = [HWND, c_void_p]
EndPaint.restype = HDC
GetObjectA = gdi32.GetObjectA
GetObjectA.argtypes = [HGDIOBJ, INT, LPVOID]
GetObjectA.restype = INT

# wrap EnumDisplayMonitors to hide the callback function:
MonitorEnumProc = WINFUNCTYPE(BOOL, HMONITOR, HDC, POINTER(RECT), LPARAM)
_EnumDisplayMonitors.argtypes = [HDC, POINTER(RECT), MonitorEnumProc, LPARAM]
_EnumDisplayMonitors.restype = BOOL


def EnumDisplayMonitors():
    results = []

    def _callback(monitor, _dc, _rect, _data):
        results.append(monitor)
        return 1

    callback = MonitorEnumProc(_callback)
    _EnumDisplayMonitors(0, None, callback, 0)
    return results


def GetIntSystemParametersInfo(key):
    rv = INT()
    r = SystemParametersInfoA(key, 0, byref(rv), 0)
    if r == 0:
        return None
    return rv.value


class WNDCLASSEX(Structure):
    _fields_ = [
        ("cbSize", UINT),
        ("style", UINT),
        ("lpfnWndProc", WNDPROC),
        ("cbClsExtra", INT),
        ("cbWndExtra", INT),
        ("hInstance", HANDLE),
        ("hIcon", HANDLE),
        ("hCursor", HANDLE),
        ("hbrBackground", HANDLE),
        ("lpszMenuName", LPCWSTR),
        ("lpszClassName", LPCWSTR),
        ("hIconSm", HANDLE),
    ]


# GUID = c_ubyte * 16


class GUID(Structure):
    # noinspection PyTypeChecker
    _fields_ = [
        ('Data1', c_ulong),
        ('Data2', c_ushort),
        ('Data3', c_ushort),
        ('Data4', c_ubyte * 8),
    ]

    def __str__(self):
        return "{%08x-%04x-%04x-%s-%s}" % (
            self.Data1,
            self.Data2,
            self.Data3,
            ''.join(["%02x" % d for d in self.Data4[:2]]),
            ''.join(["%02x" % d for d in self.Data4[2:]]),
        )


IID = GUID
REFIID = POINTER(IID)

ERROR_PIPE_NOT_CONNECTED = 233
ERROR_MORE_DATA = 234
ERROR_BROKEN_PIPE = 109
ERROR_NO_DATA = 232
ERROR_HANDLE_EOF = 38
ERROR_IO_INCOMPLETE = 996
ERROR_IO_PENDING = 997
ERROR_CANCELLED = 1223
ERROR_ACCESS_DENIED = 5
ERROR_INVALID_HANDLE = 6
ERROR_OPERATION_ABORTED = 995
ERROR_INVALID_PARAMETER = 87
ERROR_SUCCESS = 0
ERROR_COUNTER_TIMEOUT = 1121
ERROR_PIPE_BUSY = 231

IO_ERROR_STR = {
    ERROR_PIPE_NOT_CONNECTED: "PIPE_NOT_CONNECTED",
    ERROR_MORE_DATA: "MORE_DATA",
    ERROR_BROKEN_PIPE: "BROKEN_PIPE",
    ERROR_NO_DATA: "NO_DATA",
    ERROR_HANDLE_EOF: "HANDLE_EOF",
    ERROR_IO_INCOMPLETE: "IO_INCOMPLETE",
    ERROR_IO_PENDING: "IO_PENDING",
    ERROR_CANCELLED: "CANCELLED",
    ERROR_ACCESS_DENIED: "ACCESS_DENIED",
    ERROR_INVALID_HANDLE: "INVALID_HANDLE",
    ERROR_OPERATION_ABORTED: "OPERATION_ABORTED",
    ERROR_INVALID_PARAMETER: "INVALID_PARAMETER",
    ERROR_SUCCESS: "SUCCESS",
    ERROR_COUNTER_TIMEOUT: "COUNTER_TIMEOUT",
    ERROR_PIPE_BUSY: "PIPE_BUSY",
}

# https://gist.github.com/EBNull/6135237
LANG_NEUTRAL = 0x00
SUBLANG_NEUTRAL = 0x00
SUBLANG_DEFAULT = 0x01

LANG_ENGLISH = 0x09
SUBLANG_ENGLISH_US = 0x01


def MAKELANGID(primary, sublang) -> int:
    return (primary & 0xFF) | (sublang & 0xFF) << 16


LCID_ENGLISH = MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US)
LCID_DEFAULT = MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT)
LCID_NEUTRAL = MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL)


def FormatMessageSystem(message_id, langid: int = LCID_ENGLISH) -> str:
    from xpra.platform.win32.constants import (
        FORMAT_MESSAGE_ALLOCATE_BUFFER, FORMAT_MESSAGE_FROM_SYSTEM, FORMAT_MESSAGE_IGNORE_INSERTS,
    )
    sys_flag = FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS
    bufptr = LPWSTR()
    chars = kernel32.FormatMessageW(sys_flag, None, message_id, langid, byref(bufptr), 0, None)
    if not chars:
        chars = FormatMessageW(sys_flag, None, message_id, LCID_NEUTRAL, byref(bufptr), 0, None)
        if not chars:
            return str(message_id)
    if not bufptr.value:
        return str(message_id)
    val = bufptr.value[:chars]
    LocalFree(bufptr)
    return str(val)



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Sequence
from contextlib import AbstractContextManager
from ctypes import c_int, c_void_p, byref, cast, POINTER

from xpra.os_util import gi_import
from xpra.util.env import envbool, envfloat, numpy_import_context
from xpra.opengl.check import check_PyOpenGL_support
from xpra.x11.bindings.display_source import get_display_ptr
from xpra.gtk.error import xsync
from xpra.gtk.window import set_visual
from xpra.log import Logger

with numpy_import_context("OpenGL: glx context", True):
    from OpenGL import GLX
    from OpenGL.GL import GL_VENDOR, GL_RENDERER, glGetString
    from OpenGL.raw.GLX._types import struct__XDisplay, struct___GLXcontextRec

log = Logger("opengl")

ARB_CONTEXT = envbool("XPRA_OPENGL_ARB_CONTEXT", True)
CORE_PROFILE = envbool("XPRA_OPENGL_CORE_PROFILE", True)
DOUBLE_BUFFERED = envbool("XPRA_OPENGL_DOUBLE_BUFFERED", True)
SCALE_FACTOR = envfloat("XPRA_OPENGL_SCALE_FACTOR", 1)
if SCALE_FACTOR <= 0 or SCALE_FACTOR > 10:
    raise ValueError(f"invalid scale factor {SCALE_FACTOR}")

GLX_ATTRIBUTES: dict[Any, str] = {
    GLX.GLX_ACCUM_RED_SIZE: "accum-red-size",
    GLX.GLX_ACCUM_GREEN_SIZE: "accum-green-size",
    GLX.GLX_ACCUM_BLUE_SIZE: "accum-blue-size",
    GLX.GLX_ACCUM_ALPHA_SIZE: "accum-alpha-size",
    GLX.GLX_RED_SIZE: "red-size",
    GLX.GLX_GREEN_SIZE: "green-size",
    GLX.GLX_BLUE_SIZE: "blue-size",
    GLX.GLX_ALPHA_SIZE: "alpha-size",
    GLX.GLX_DEPTH_SIZE: "depth-size",
    GLX.GLX_STENCIL_SIZE: "stencil-size",
    GLX.GLX_BUFFER_SIZE: "buffer-size",
    GLX.GLX_AUX_BUFFERS: "aux-buffers",
    GLX.GLX_DOUBLEBUFFER: "double-buffered",
    GLX.GLX_LEVEL: "level",
    GLX.GLX_STEREO: "stereo",
    GLX.GLX_RGBA: "rgba",
}


def c_attrs(props: dict):
    attrs = []
    for k, v in props.items():
        if v is None:
            attrs += [k]
        else:
            attrs += [k, v]
    attrs += [0, 0]
    # noinspection PyTypeChecker,PyCallingNonCallable
    return (c_int * len(attrs))(*attrs)


XDISPLAY = int    # POINTER(struct__XDisplay)


def get_xdisplay() -> XDISPLAY:
    ptr = get_display_ptr()
    if not ptr:
        raise RuntimeError("no X11 display registered")
    return cast(ptr, POINTER(struct__XDisplay))


def get_extensions(xdisplay: int) -> Sequence[str]:
    bext = GLX.glXQueryExtensionsString(xdisplay, 0)
    if not bext:
        return ()
    str_ext = bext.decode("latin1")
    return tuple(x for x in str_ext.strip().split(" ") if x)


def get_fbconfig_attributes(xdisplay: int, fbconfig) -> dict[str, int]:
    fb_attrs: dict[str, int] = {}
    for name, attr in {
        "fbconfig-id": GLX.GLX_FBCONFIG_ID,
        "level": GLX.GLX_LEVEL,
        "double-buffer": GLX.GLX_DOUBLEBUFFER,
        "stereo": GLX.GLX_STEREO,
        "aux-buffers": GLX.GLX_AUX_BUFFERS,
        "red-size": GLX.GLX_RED_SIZE,
        "green-size": GLX.GLX_GREEN_SIZE,
        "blue-size": GLX.GLX_BLUE_SIZE,
        "alpha-size": GLX.GLX_ALPHA_SIZE,
        "depth-size": GLX.GLX_DEPTH_SIZE,
        "stencil-size": GLX.GLX_STENCIL_SIZE,
        "accum-red-size": GLX.GLX_ACCUM_RED_SIZE,
        "accum-green-size": GLX.GLX_ACCUM_GREEN_SIZE,
        "accum-blue-size": GLX.GLX_ACCUM_BLUE_SIZE,
        "accum-alpha-size": GLX.GLX_ACCUM_ALPHA_SIZE,
        "render-type": GLX.GLX_RENDER_TYPE,
        "drawable-type": GLX.GLX_DRAWABLE_TYPE,
        "renderable": GLX.GLX_X_RENDERABLE,
        "visual-id": GLX.GLX_VISUAL_ID,
        "visual-type": GLX.GLX_X_VISUAL_TYPE,
        "config-caveat": GLX.GLX_CONFIG_CAVEAT,
        "transparent-type": GLX.GLX_TRANSPARENT_TYPE,
        "transparent-index-value": GLX.GLX_TRANSPARENT_INDEX_VALUE,
        "transparent-red-value": GLX.GLX_TRANSPARENT_RED_VALUE,
        "transparent-green-value": GLX.GLX_TRANSPARENT_GREEN_VALUE,
        "transparent-blue-value": GLX.GLX_TRANSPARENT_BLUE_VALUE,
        "transparent-alpha-value": GLX.GLX_TRANSPARENT_ALPHA_VALUE,
        "max-pbuffer-width": GLX.GLX_MAX_PBUFFER_WIDTH,
        "max-pbuffer-height": GLX.GLX_MAX_PBUFFER_HEIGHT,
        "max-pbuffer-pixels": GLX.GLX_MAX_PBUFFER_PIXELS,
    }.items():
        value = c_int()
        if not GLX.glXGetFBConfigAttrib(xdisplay, fbconfig, attr, byref(value)):
            fb_attrs[name] = value.value
    return fb_attrs


class GLXWindowContext(AbstractContextManager):

    def __init__(self, glx_context, xid: int):
        self.context = glx_context
        self.xid = xid
        self.xdisplay: XDISPLAY = get_xdisplay()
        self.valid: bool = False

    def __enter__(self):
        log("glXMakeCurrent: xid=%#x, context=%s", self.xid, self.context)
        with xsync:
            if not GLX.glXMakeCurrent(self.xdisplay, self.xid, self.context):
                raise RuntimeError("glXMakeCurrent failed")
        self.valid = True
        return self

    def __exit__(self, *_args):
        self.valid = False
        if self.context:
            context_type = type(self.context)
            null_context = cast(0, context_type)
            log("glXMakeCurrent: NULL for xid=%#x", self.xid)
            if not GLX.glXMakeCurrent(self.xdisplay, 0, null_context):
                log.error("Error: glXMakeCurrent NULL failed")

    def update_geometry(self) -> None:
        """ not needed on X11 """

    def swap_buffers(self) -> None:
        assert self.valid, "GLX window context is no longer valid"
        GLX.glXSwapBuffers(self.xdisplay, self.xid)

    def get_scale_factor(self) -> float:
        return SCALE_FACTOR

    def __repr__(self):
        return "GLXWindowContext(%#x)" % self.xid


class GLXContext:

    def __init__(self, alpha=False):
        self.props: dict[str, Any] = {}
        self.xdisplay: int = 0
        self.context = None
        self.bit_depth: int = 0
        self.xdisplay = get_xdisplay()

        # query version
        major = c_int()
        minor = c_int()
        if not GLX.glXQueryVersion(self.xdisplay, byref(major), byref(minor)):
            raise RuntimeError("failed to query GLX version")
        log("found GLX version %i.%i", major.value, minor.value)
        self.props["GLX"] = (major.value, minor.value)

        # find a framebuffer config we can use:
        bpc = 8
        pyattrs = {
            GLX.GLX_X_RENDERABLE: True,
            GLX.GLX_DRAWABLE_TYPE: GLX.GLX_WINDOW_BIT,
            GLX.GLX_RENDER_TYPE: GLX.GLX_RGBA_BIT,
            GLX.GLX_X_VISUAL_TYPE: GLX.GLX_TRUE_COLOR,
            GLX.GLX_RED_SIZE: bpc,
            GLX.GLX_GREEN_SIZE: bpc,
            GLX.GLX_BLUE_SIZE: bpc,
            GLX.GLX_DEPTH_SIZE: 24,
            GLX.GLX_STENCIL_SIZE: 8,
        }
        if alpha:
            pyattrs[GLX.GLX_ALPHA_SIZE] = int(alpha) * bpc
        if DOUBLE_BUFFERED:
            pyattrs[GLX.GLX_DOUBLEBUFFER] = True
        attrs = c_attrs(pyattrs)
        fbcount = c_int()
        fbc = GLX.glXChooseFBConfig(self.xdisplay, 0, attrs, byref(fbcount))
        log(f"glXChooseFBConfig(..)={fbc} {fbcount=}")
        if fbcount.value <= 0:
            raise RuntimeError(f"no frame buffer configurations found matching {pyattrs}")
        for i in range(fbcount.value):
            fb_attrs = get_fbconfig_attributes(self.xdisplay, fbc[i])
            log(f"[{i:2}] {fb_attrs}")
        fbconfig = fbc[0]
        log(f"using {fbconfig=}")
        # the X11 visual for this framebuffer config:
        xvinfo = GLX.glXGetVisualFromFBConfig(self.xdisplay, fbconfig)

        # query extensions:
        extensions = get_extensions(self.xdisplay)
        log(f"{extensions=}")

        def getconfig(attr: int) -> int:
            value = c_int()
            r = GLX.glXGetConfig(self.xdisplay, xvinfo, attr, byref(value))
            if r:
                raise RuntimeError(f"glXGetConfig({attr}) returned {r}")
            return value.value

        if not getconfig(GLX.GLX_USE_GL):
            raise RuntimeError("OpenGL is not supported by this visual!")

        self.bit_depth = sum(getconfig(x) for x in (
            GLX.GLX_RED_SIZE,
            GLX.GLX_GREEN_SIZE,
            GLX.GLX_BLUE_SIZE,
            GLX.GLX_ALPHA_SIZE,
        ))
        self.props["depth"] = self.bit_depth
        # hide those because we don't use them
        # and because they're misleading: 'has-alpha' may be False
        # even when we have RGBA support (and therefore very likely to have alpha..)
        # self.props["has-depth-buffer"] = getconfig(GLX.GLX_DEPTH_SIZE)>0
        # self.props["has-stencil-buffer"] = getconfig(GLX.GLX_STENCIL_SIZE)>0
        # self.props["has-alpha"] = getconfig(GLX.GLX_ALPHA_SIZE)>0
        for attrib, name in GLX_ATTRIBUTES.items():
            v = getconfig(attrib)
            if name in ("stereo", "double-buffered", "rgba"):
                v = bool(v)
            self.props[name] = v
        # attribute names matching gtkgl:
        display_mode = []
        if getconfig(GLX.GLX_RGBA):
            # this particular context may not have alpha channel support...
            # but if we have RGBA then it's almost guaranteed that we can do ALPHA
            display_mode.append("ALPHA")
        if getconfig(GLX.GLX_DOUBLEBUFFER):
            display_mode.append("DOUBLE")
        else:  # pragma: no cover
            display_mode.append("SINGLE")
        self.props["display_mode"] = display_mode
        if ARB_CONTEXT and "GLX_ARB_create_context" in extensions:
            from OpenGL.raw.GLX.ARB.create_context import (
                # glXCreateContextAttribsARB,
                GLX_CONTEXT_FLAGS_ARB,
                GLX_CONTEXT_MAJOR_VERSION_ARB, GLX_CONTEXT_MINOR_VERSION_ARB,
            )
            from OpenGL.raw.GLX.ARB.create_context_profile import (
                GLX_CONTEXT_PROFILE_MASK_ARB,
                GLX_CONTEXT_CORE_PROFILE_BIT_ARB, GLX_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB,
            )
            if CORE_PROFILE:
                profile_mask = GLX_CONTEXT_CORE_PROFILE_BIT_ARB
            else:
                profile_mask = GLX_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB
            context_attrs = c_attrs({
                GLX_CONTEXT_MAJOR_VERSION_ARB: 3,
                GLX_CONTEXT_MINOR_VERSION_ARB: 2,
                GLX_CONTEXT_PROFILE_MASK_ARB: profile_mask,
                GLX_CONTEXT_FLAGS_ARB: 0,
            })
            glXCreateContextAttribsARB = GLX.glXGetProcAddress("glXCreateContextAttribsARB")
            log(f"{glXCreateContextAttribsARB=} {context_attrs=}")
            glXCreateContextAttribsARB.argtypes = [POINTER(struct__XDisplay), c_void_p, c_int, c_int, POINTER(c_int)]
            glXCreateContextAttribsARB.restype = POINTER(struct___GLXcontextRec)
            self.context = glXCreateContextAttribsARB(self.xdisplay, fbconfig, 0, True, context_attrs)
        else:
            self.context = GLX.glXCreateNewContext(self.xdisplay, fbconfig, GLX.GLX_RGBA_TYPE, None, True)
        log(f"gl context={self.context}")
        self.props["direct"] = bool(GLX.glXIsDirect(self.xdisplay, self.context))

        def getstr(k) -> str:
            try:
                return glGetString(k)
            except Exception as e:  # pragma: no cover
                self.props["safe"] = False
                result = getattr(e, "result", None)
                if result and isinstance(result, str):
                    return result
                raise

        self.props["vendor"] = getstr(GL_VENDOR)
        self.props["renderer"] = getstr(GL_RENDERER)
        log("GLXContext(%s) context=%s, props=%s", alpha, self.context, self.props)

    def check_support(self, force_enable=False) -> dict[str, Any]:
        i = self.props
        if not self.xdisplay:
            return {
                "success": False,
                "safe": False,
                "enabled": False,
                "message": "cannot access X11 display",
            }
        Gtk = gi_import("Gtk")
        tmp = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
        tmp.resize(1, 1)
        tmp.set_decorated(False)
        tmp.realize()
        da = Gtk.DrawingArea()
        tmp.add(da)
        set_visual(da, True)
        win = tmp.get_window()
        log("check_support(%s) using temporary window=%s", force_enable, tmp)
        with self.get_paint_context(win):
            i.update(check_PyOpenGL_support(force_enable))
        tmp.destroy()
        return i

    def get_bit_depth(self) -> int:
        return self.bit_depth

    def is_double_buffered(self) -> bool:
        return DOUBLE_BUFFERED

    def get_paint_context(self, gdk_window) -> GLXWindowContext:
        if not self.context:
            raise RuntimeError("no glx context")
        if not gdk_window:
            raise RuntimeError("cannot get a paint context without a window")
        return GLXWindowContext(self.context, gdk_window.get_xid())

    def destroy(self) -> None:
        c = self.context
        if c:
            self.context = None
            GLX.glXDestroyContext(self.xdisplay, c)

    def __repr__(self):
        return f"GLXContext({self.props})"


GLContext = GLXContext


def check_support() -> dict[str, Any]:
    ptr = get_display_ptr()
    if not ptr:
        from xpra.x11.gtk.display_source import init_gdk_display_source
        init_gdk_display_source()

    return GLContext().check_support()



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os


def get_sys_info() -> dict:
    return {}


def get_name() -> str:
    try:
        from ctypes import (
            WinError, get_last_error,  # @UnresolvedImport
            c_char, byref,
        )
        from ctypes.wintypes import DWORD
        from xpra.platform.win32.common import GetUserNameA
        max_len = 256
        size = DWORD(max_len)
        # noinspection PyTypeChecker,PyCallingNonCallable
        buf = (c_char * (max_len + 1))()
        if not GetUserNameA(byref(buf), byref(size)):
            raise WinError(get_last_error())
        return buf.value[:size.value].decode()
    except Exception:
        return os.environ.get("USERNAME", "")


def get_version_info() -> dict:
    return {}



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest
from time import monotonic
from collections.abc import Callable

from xpra.util.objects import typedict, AdHocStruct
from xpra.log import Logger
from xpra.os_util import gi_import

GLib = gi_import("GLib")


def debug_all() -> None:
    from xpra.log import enable_debug_for
    enable_debug_for("all")


class ClientMixinTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super(ClientMixinTest, cls).setUpClass()
        cls.glib = GLib
        cls.main_loop = cls.glib.MainLoop()

    def setUp(self):
        self.packets = []
        self.mixin = None
        self.packet_handlers = {}
        self.exit_codes = []
        self.legacy_alias = {}

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        if self.mixin:
            self.mixin.cleanup()
            self.mixin = None

    def stop(self) -> None:
        self.glib.timeout_add(1000, self.main_loop.quit)

    def dump_packets(self) -> None:
        from xpra.util.io import get_util_logger
        log = get_util_logger()
        log.info("dump_packets() %i packets to send:", len(self.packets))
        for x in self.packets:
            log.info("%s", x)

    def send(self, *args) -> None:
        self.packets.append(args)

    def get_packet(self, index: int):
        if index < 0:
            actual_index = len(self.packets)+index
        else:
            actual_index = index
        assert actual_index >= 0, "invalid actual index %i for index %i" % (actual_index, index)
        assert len(self.packets) > actual_index, "not enough packets (%i) to access %i" % (len(self.packets), index)
        return self.packets[actual_index]

    def verify_packet(self, index: int, expected) -> None:
        packet = self.get_packet(index)
        pslice = packet[:len(expected)]
        if pslice != expected:
            log = Logger("test")
            log.error(f"packet mismatch at index {index}")
            for i, packet in enumerate(self.packets):
                log.error("[%3i] %s", i, packet)
            raise RuntimeError("invalid packet slice %s, expected %s" % (pslice, expected))

    def add_legacy_alias(self, legacy_name: str, name: str) -> None:
        self.legacy_alias[legacy_name] = name

    def add_packet_handler(self, packet_type: str, handler: Callable, main_thread=False):
        # log("add_packet_handler%s", (packet_type, handler, main_thread))
        self.packet_handlers[packet_type] = handler

    def add_packets(self, *packet_types: str, main_thread=False) -> None:
        for packet_type in packet_types:
            handler = getattr(self.mixin, "_process_" + packet_type.replace("-", "_"))
            self.add_packet_handler(packet_type, handler, main_thread)

    def handle_packet(self, packet) -> None:
        packet_type = packet[0]
        ph = self.packet_handlers.get(packet_type)
        assert ph is not None, "no packet handler for %s" % packet_type
        ph(packet)

    def fake_quit(self, code) -> None:
        self.exit_codes.append(code)

    def _test_mixin_class(self, mclass, opts, caps=None, protocol_type="xpra"):
        x = self.mixin = mclass()
        x.start_time = monotonic()
        x.exit_code = None
        x.server_packet_types = ()
        x.quit = self.fake_quit
        fake_protocol = AdHocStruct()
        fake_protocol.get_info = lambda: {}
        fake_protocol.set_compression_level = lambda _x: None
        fake_protocol.TYPE = protocol_type
        x._protocol = fake_protocol   # pylint: disable=protected-access
        x.add_legacy_alias = self.add_legacy_alias
        x.add_packets = self.add_packets
        x.add_packet_handler = self.add_packet_handler
        x.init(opts)
        conn = AdHocStruct()
        conn.filename = "/fake/path/to/nowhere"
        x.setup_connection(conn)
        x.send = self.send
        x.send_now = self.send
        x.init_authenticated_packet_handlers()
        caps = self.make_caps(caps)
        x.parse_server_capabilities(caps)
        x.process_ui_capabilities(caps)
        assert x.get_caps() is not None
        assert x.get_info() is not None
        return x

    def make_caps(self, caps=None) -> typedict:
        return typedict(caps or {})



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.util.str_fn import print_nested_dict
from xpra.x11.gtk.display_source import init_gdk_display_source
from xpra.x11.bindings.randr import RandRBindings  # pylint: disable=no-name-in-module


def main() -> None:
    init_gdk_display_source()
    randr = RandRBindings()
    # print(randr.is_dummy16())
    print_nested_dict(randr.get_all_screen_properties())


if __name__ == "__main__":
    sys.exit(main())



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from subprocess import check_call

from xpra.os_util import POSIX, gi_import
from xpra.log import Logger

log = Logger("util")


def sync() -> None:
    if POSIX:
        check_call("sync")


def run_gui(gui_class) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    from xpra.log import enable_color
    from xpra.platform.gui import init, ready
    from xpra.gtk.signals import install_signal_handlers
    with program_context("xpra-configure-gui", "Xpra Configure GUI"):
        enable_color()
        init()
        gui = gui_class()
        install_signal_handlers("xpra-configure-gui", gui.app_signal)
        ready()
        gui.show()
        Gtk = gi_import("Gtk")
        Gtk.main()
        log("do_main() gui.exit_code=%i", gui.exit_code)
        return 0



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.net import packet_encoding

class TestPacketEncoding(unittest.TestCase):

    def test_env_log(self):
        packet_encoding.init_all()
        assert packet_encoding.get_packet_encoding_caps()
        assert packet_encoding.get_enabled_encoders()
        for x in packet_encoding.get_enabled_encoders():
            e = packet_encoding.get_encoder(x)
            for packet_data in (
                ["hello", {"foo" : 1, "bar" : True}],
                #b"foo",
                ):
                assert e
                v, flag = e(packet_data)
                assert v
                if x=="none":
                    #'none' cannot decode back
                    continue
                try:
                    r = packet_encoding.decode(v, flag)
                    assert r
                except Exception:
                    print("error calling decode(%s, %s) for encoder %s" % (v, flag, x))
                    raise
        #one-shot function:
        packet = ("hello", {})
        assert packet_encoding.pack_one_packet(packet)

def main():
    unittest.main()

if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.os_util import gi_import
from xpra.client.base.client import XpraClientBase, EXTRA_TIMEOUT
from xpra.exit_codes import ExitCode, ExitValue
from xpra.log import Logger

log = Logger("gobject", "client")

GObject = gi_import("GObject")
GLib = gi_import("GLib")


class GObjectXpraClient(GObject.GObject, XpraClientBase):
    """
        Utility superclass for GObject clients
    """
    COMMAND_TIMEOUT = EXTRA_TIMEOUT

    def __init__(self):
        self.glib_mainloop = None
        GObject.GObject.__init__(self)
        XpraClientBase.__init__(self)

    def init(self, opts) -> None:
        XpraClientBase.init(self, opts)
        self.glib_init()

    def install_signal_handlers(self) -> None:
        from xpra.gtk.signals import install_signal_handlers
        install_signal_handlers("%s Client" % self.client_type(), self.handle_app_signal)

    def make_protocol(self, conn):
        protocol = super().make_protocol(conn)
        protocol._log_stats = False
        GLib.idle_add(self.send_hello)
        return protocol

    def client_type(self) -> str:
        # overridden in subclasses!
        return "pygobject"

    def run(self) -> ExitValue:
        XpraClientBase.run(self)
        self.run_loop()
        return self.exit_code or 0

    def run_loop(self) -> None:
        self.glib_mainloop = GLib.MainLoop()
        self.glib_mainloop.run()

    def make_hello(self) -> dict[str, Any]:
        capabilities = XpraClientBase.make_hello(self)
        capabilities["keyboard"] = False
        return capabilities

    def quit(self, exit_code: int | ExitCode = 0) -> None:
        log("quit(%s) current exit_code=%s", exit_code, self.exit_code)
        if self.exit_code is None:
            self.exit_code = exit_code
        self.cleanup()
        GLib.timeout_add(50, self.exit_loop)

    def exit_loop(self) -> None:
        self.glib_mainloop.quit()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False
from time import monotonic
from typing import Tuple, NoReturn

from xpra.log import Logger
log = Logger("decoder", "jpeg")

from xpra.util.env import envbool
from xpra.util.objects import reverse_dict, typedict
from xpra.codecs.image import ImageWrapper
from xpra.buffers.membuf cimport getbuf, MemBuf  # pylint: disable=syntax-error
from libc.stdint cimport uintptr_t, uint8_t
from libc.string cimport memset  # pylint: disable=syntax-error

LOG_PERF = envbool("XPRA_JPEG_LOG_PERF", False)

ctypedef int TJSAMP
ctypedef int TJPF
ctypedef int TJCS

cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS

cdef extern from "turbojpeg.h":
    TJSAMP  TJSAMP_444
    TJSAMP  TJSAMP_422
    TJSAMP  TJSAMP_420
    TJSAMP  TJSAMP_GRAY
    TJSAMP  TJSAMP_440
    TJSAMP  TJSAMP_411

    TJPF    TJPF_RGB
    TJPF    TJPF_BGR
    TJPF    TJPF_RGBX
    TJPF    TJPF_BGRX
    TJPF    TJPF_XBGR
    TJPF    TJPF_XRGB
    TJPF    TJPF_GRAY
    TJPF    TJPF_RGBA
    TJPF    TJPF_BGRA
    TJPF    TJPF_ABGR
    TJPF    TJPF_ARGB
    TJPF    TJPF_CMYK

    TJCS    TJCS_RGB
    TJCS    TJCS_YCbCr
    TJCS    TJCS_GRAY
    TJCS    TJCS_CMYK
    TJCS    TJCS_YCCK

    int TJFLAG_BOTTOMUP
    int TJFLAG_FASTUPSAMPLE
    int TJFLAG_FASTDCT
    int TJFLAG_ACCURATEDCT

    ctypedef void* tjhandle
    tjhandle tjInitDecompress()
    int tjDecompressHeader3(tjhandle handle,
                            const unsigned char *jpegBuf, unsigned long jpegSize, int *width,
                            int *height, int *jpegSubsamp, int *jpegColorspace)
    int tjDecompress2(tjhandle handle,
                      const unsigned char *jpegBuf, unsigned long jpegSize, unsigned char *dstBuf,
                      int width, int pitch, int height, int pixelFormat, int flags) nogil
    int tjDecompressToYUVPlanes(tjhandle handle,
                                const unsigned char *jpegBuf, unsigned long jpegSize,
                                unsigned char **dstPlanes, int width, int *strides, int height, int flags) nogil
    int tjDestroy(tjhandle handle)
    char* tjGetErrorStr()

    int tjPlaneWidth(int componentID, int width, int subsamp)
    unsigned long tjBufSizeYUV2(int width, int pad, int height, int subsamp)
    unsigned long tjPlaneSizeYUV(int componentID, int width, int stride, int height, int subsamp)
    int tjPlaneWidth(int componentID, int width, int subsamp)
    int tjPlaneHeight(int componentID, int height, int subsamp)


TJSAMP_STR = {
    TJSAMP_444  : "444",
    TJSAMP_422  : "422",
    TJSAMP_420  : "420",
    TJSAMP_GRAY : "GRAY",
    TJSAMP_440  : "440",
    TJSAMP_411  : "411",
}

TJCS_STR = {
    TJCS_RGB    : "RGB",
    TJCS_YCbCr  : "YCbCr",
    TJCS_GRAY   : "GRAY",
    TJCS_CMYK   : "CMYK",
    TJCS_YCCK   : "YCCK",
}

TJPF_STR = {
    TJPF_RGB    : "RGB",
    TJPF_BGR    : "BGR",
    TJPF_RGBX   : "RGBX",
    TJPF_BGRX   : "BGRX",
    TJPF_XBGR   : "XBGR",
    TJPF_XRGB   : "XRGB",
    TJPF_GRAY   : "GRAY",
    TJPF_RGBA   : "RGBA",
    TJPF_BGRA   : "BGRA",
    TJPF_ABGR   : "ABGR",
    TJPF_ARGB   : "ARGB",
    TJPF_CMYK   : "CMYK",
}
TJPF_VAL = reverse_dict(TJPF_STR)


def get_version() -> Tuple[int, int]:
    return (1, 0)


def get_encodings() -> Sequence[str]:
    return ("jpeg", "jpega")


cdef inline int roundup(int n, int m):
    return (n + m - 1) & ~(m - 1)

DEF ALIGN = 4


def get_error_str() -> str:
    cdef char *err = tjGetErrorStr()
    return str(err)


cdef inline void close_handle(tjhandle handle):
    r = tjDestroy(handle)
    if r:
        log.error(f"Error: failed to destroy the JPEG decompressor, code {r}:")
        log.error(" %s", get_error_str())


def tj_err(int r, msg: str) -> NoReturn:
    err = get_error_str()
    raise RuntimeError(f"{msg}: {err!r}")


def decompress_to_yuv(data: bytes, options: typedict) -> ImageWrapper:
    cdef tjhandle decompressor = tjInitDecompress()
    if decompressor==NULL:
        raise RuntimeError("failed to instantiate a JPEG decompressor")

    cdef Py_buffer py_buf
    if PyObject_GetBuffer(data, &py_buf, PyBUF_ANY_CONTIGUOUS):
        tjDestroy(decompressor)
        raise RuntimeError(f"failed to read compressed data from {type(data)}")

    cdef int r

    def close_decompressor() -> None:
        close_handle(decompressor)

    def release() -> None:
        PyBuffer_Release(&py_buf)

    def close() -> None:
        release()
        close_decompressor()

    def tj_check(int r, msg: str):
        if r:
            close()
            tj_err(r, msg)

    # set the buffer size and pointer to `jpeg` portion:
    cdef unsigned int alpha_offset = options.intget("alpha-offset", 0)
    cdef unsigned char nplanes = 3
    cdef unsigned int buf_len = py_buf.len
    cdef void* buf = <void*> py_buf.buf
    if alpha_offset:
        if alpha_offset >= buf_len:
            raise ValueError(f"jpeg data buffer is too small, expected at least {alpha_offset} but got {buf_len}")
        buf_len = alpha_offset
        nplanes = 4

    # parse header:
    cdef int w, h, subsamp, cs
    r = tjDecompressHeader3(decompressor,
                            <const unsigned char *> buf, buf_len,
                            &w, &h, &subsamp, &cs)
    tj_check(r, "failed to decompress JPEG main header")

    subsamp_str = TJSAMP_STR.get(subsamp, subsamp)
    assert subsamp in (TJSAMP_444, TJSAMP_422, TJSAMP_420, TJSAMP_GRAY), "unsupported JPEG colour subsampling: %s" % subsamp_str
    log("jpeg.decompress_to_yuv %i planes, size: %4ix%-4i, subsampling=%-4s, colorspace=%s",
        nplanes, w, h, subsamp_str, TJCS_STR.get(cs, cs))
    if nplanes==3:
        pixel_format = "YUV%sP" % subsamp_str
    elif nplanes == 4:
        pixel_format = "YUVA%sP" % subsamp_str
    elif nplanes==1:
        pixel_format = "YUV400P"
    else:
        close()
        raise ValueError("invalid number of planes: %i" % nplanes)
    #allocate YUV buffers:
    cdef unsigned long plane_size
    cdef unsigned char *planes[4]
    cdef int strides[4]
    cdef int i, stride
    cdef MemBuf membuf
    cdef MemBuf empty
    cdef int flags = 0
    pystrides = []
    pyplanes = []
    cdef unsigned long total_size = 0
    cdef double start, elapsed
    for i in range(4):
        strides[i] = 0
        planes[i] = NULL

    for i in range(nplanes):
        if i == 3:
            subsamp = TJSAMP_GRAY
        stride = tjPlaneWidth(i % 3, w, subsamp)
        if stride <= 0:
            if subsamp != TJSAMP_GRAY or (i % 3) == 0:
                raise ValueError("cannot get size for plane %r for mode %r" % ("YUVA"[i], subsamp_str))
            stride = roundup(w//2, ALIGN)
            plane_size = stride * roundup(h, 2)//2
            if i==1:
                #allocate empty U and V planes:
                empty = getbuf(plane_size, 0)
                memset(<void *> empty.get_mem(), 128, plane_size)
                pixel_format = "YUV420P"
            membuf = empty
        else:
            stride = roundup(stride, ALIGN)
            strides[i] = stride
            plane_size = tjPlaneSizeYUV(i % 3, w, stride, h, subsamp)
            membuf = getbuf(plane_size, 0)     #add padding?
            planes[i] = <unsigned char*> membuf.get_mem()
        total_size += plane_size
        #python objects for each plane:
        pystrides.append(stride)
        pyplanes.append(memoryview(membuf))
    # log("jpeg strides: %s, plane sizes=%s", pystrides, [int(plane_sizes[i]) for i in range(4)])
    start = monotonic()
    with nogil:
        r = tjDecompressToYUVPlanes(decompressor,
                                    <const unsigned char*> buf, buf_len,
                                    planes, w, strides, h, flags)
    tj_check(r, "failed to decompress {subsamp_str!r} JPEG data to YUV")

    cdef int aw, ah
    if alpha_offset:
        # now decompress the alpha channel:
        buf = <void*> (<uintptr_t> py_buf.buf + alpha_offset)
        buf_len = py_buf.len - alpha_offset

        r = tjDecompressHeader3(decompressor,
                                <const unsigned char *> buf, buf_len,
                                &aw, &ah, &subsamp, &cs)
        tj_check(r, "failed to decompress JPEG alpha header")

        if subsamp != TJSAMP_GRAY:
            subsamp_str = TJSAMP_STR.get(subsamp, subsamp)
            raise ValueError("unsupported JPEG colour subsampling for alpha channel: %s" % subsamp_str)

        if aw != w or ah != h:
            raise ValueError(f"alpha channel dimensions {aw}x{ah} does not match main image {w}x{h}")

        with nogil:
            r = tjDecompressToYUVPlanes(decompressor,
                                        <const unsigned char*> buf, buf_len,
                                        &planes[3], aw, &strides[3], ah, flags)
        tj_check(r, "failed to decompress JPEG alpha channel data")

    close_decompressor()
    release()

    if LOG_PERF:
        elapsed = monotonic()-start
        log("decompress jpeg to %s: %4i MB/s (%9i bytes in %2.1fms)",
            pixel_format, total_size/elapsed//1024//1024, total_size, 1000*elapsed)
    return ImageWrapper(0, 0, w, h, pyplanes, pixel_format, nplanes*8, pystrides, planes=nplanes)


def decompress_to_rgb(data: bytes, options: typedict) -> ImageWrapper:
    cdef unsigned int alpha_offset = options.intget("alpha-offset", 0)
    rgb_format = "BGRA" if alpha_offset else "BGRX"
    rgb_format = options.strget("rgb_format", rgb_format)
    assert rgb_format in TJPF_VAL
    cdef TJPF pixel_format = TJPF_VAL[rgb_format]

    cdef tjhandle decompressor = tjInitDecompress()
    if decompressor==NULL:
        raise RuntimeError("failed to instantiate a JPEG decompressor")

    cdef Py_buffer py_buf
    if PyObject_GetBuffer(data, &py_buf, PyBUF_ANY_CONTIGUOUS):
        tjDestroy(decompressor)
        raise ValueError(f"failed to read compressed data from {type(data)}")

    cdef int r

    def close_decompressor() -> None:
        close_handle(decompressor)

    def release() -> None:
        PyBuffer_Release(&py_buf)

    def close() -> None:
        release()
        close_decompressor()

    def tj_check(int r, msg: str):
        if r:
            close()
            tj_err(r, msg)

    cdef int w, h, subsamp, cs
    cdef uintptr_t buf = <uintptr_t> py_buf.buf
    cdef unsigned long buf_size = py_buf.len
    if alpha_offset>0:
        buf_size = alpha_offset
    log("decompressing buffer at %#x of size %i", buf, buf_size)
    r = tjDecompressHeader3(decompressor,
                            <const unsigned char *> buf, buf_size,
                            &w, &h, &subsamp, &cs)
    tj_check(r, "failed to decompress JPEG header")
    subsamp_str = TJSAMP_STR.get(subsamp, subsamp)
    log("jpeg.decompress_to_rgb: size=%4ix%-4i, subsampling=%3s, colorspace=%s",
        w, h, subsamp_str, TJCS_STR.get(cs, cs))
    cdef int flags = 0      #TJFLAG_BOTTOMUP
    cdef double elapsed
    cdef double start = monotonic()
    cdef int stride = w*4
    cdef unsigned long size = stride*h
    cdef MemBuf membuf = getbuf(size, 0)
    cdef unsigned char *dst_buf = <unsigned char*> membuf.get_mem()
    with nogil:
        r = tjDecompress2(decompressor,
                          <const unsigned char *> buf, buf_size, dst_buf,
                          w, stride, h, pixel_format, flags)
    tj_check(r, "failed to decompress {subsamp_str!r} JPEG data to {rgb_format!r")
    # deal with alpha channel if there is one:
    cdef int aw, ah
    cdef unsigned char *planes[3]
    cdef int strides[3]
    cdef MemBuf alpha
    cdef unsigned long alpha_size
    cdef int x, y, alpha_stride
    cdef unsigned char* alpha_plane
    cdef char alpha_index
    bpp = 24
    if alpha_offset:
        bpp = 32
        alpha_index = rgb_format.find("A")
        assert alpha_index>=0, "no 'A' in %s" % rgb_format
        assert len(rgb_format)==4, "unsupported rgb format for alpha: %s" % rgb_format
        assert <unsigned long> py_buf.len>alpha_offset, "alpha offset is beyond the end of the compressed buffer"
        buf = (<uintptr_t> py_buf.buf) + alpha_offset
        buf_len = py_buf.len - alpha_offset
        r = tjDecompressHeader3(decompressor,
                                <const unsigned char *> buf, buf_len,
                                &aw, &ah, &subsamp, &cs)
        tj_check(r, "failed to decompress alpha channel")
        assert aw==w and ah==h, "alpha plane dimensions %ix%i don't match main image %ix%i" % (aw, ah, w, h)
        subsamp_str = TJSAMP_STR.get(subsamp, subsamp)
        log("found alpha plane %r at %#x size %i", subsamp_str, buf, buf_len)
        assert subsamp==TJSAMP_GRAY, "unsupported JPEG alpha subsampling: %s" % subsamp_str
        for i in range(3):
            strides[i] = 0
            planes[i] = NULL
        alpha_stride = tjPlaneWidth(0, w, subsamp)
        strides[0] = alpha_stride
        alpha_size = tjPlaneSizeYUV(0, w, alpha_stride, h, subsamp)
        alpha = getbuf(alpha_size, 0)
        alpha_plane = <unsigned char*> alpha.get_mem()
        planes[0] = alpha_plane
        with nogil:
            r = tjDecompressToYUVPlanes(decompressor,
                                        <const unsigned char*> buf, buf_len,
                                        planes, w, strides, h, flags)
        tj_check(r, f"failed to decompress {subsamp_str!r} JPEG alpha data")
        #merge alpha into rgb buffer:
        for y in range(h):
            for x in range(w):
                dst_buf[y*stride+x*4+alpha_index] = alpha_plane[y*alpha_stride+x]

    close()
    if LOG_PERF:
        elapsed = monotonic()-start
        log("decompress jpeg to %s: %4i MB/s (%9i bytes in %2.1fms)",
            rgb_format, size/elapsed//1024//1024, size, 1000*elapsed)
    return ImageWrapper(0, 0, w, h, memoryview(membuf), rgb_format, bpp, stride, planes=ImageWrapper.PACKED)


def selftest(full=False) -> None:
    log("jpeg selftest")

    def test_rgbx(bdata, options):
        return decompress_to_rgb(bdata, typedict(options))

    def test_yuv(bdata, options):
        return decompress_to_yuv(bdata, typedict(options))

    from xpra.codecs.checks import TEST_PICTURES
    for encoding in ("jpeg", "jpega"):
        for size, samples in TEST_PICTURES[encoding].items():
            w, h = size
            for data, options in samples:
                for fn in (test_yuv, test_rgbx):
                    img = fn(data, options)
                    log("%s(%i bytes, %s)=%s", fn, len(data), options, img)
                    if full:
                        try:
                            v = fn(data[:len(data)//2])
                            assert v is not None
                        except:
                            pass
                        else:
                            raise RuntimeError("should not be able to decompress incomplete data, but got %s" % v)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic
from typing import Any
from collections.abc import Callable, Sequence

from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.env import envint
from xpra.common import NotificationID, ConnectionMessage
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("timeout")

GRACE_PERCENT = envint("XPRA_GRACE_PERCENT", 90)


class IdleMixin(StubSourceMixin):

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        return caps.boolget("keyboard") or caps.boolget("mouse") or caps.boolget("windows")

    def __init__(self) -> None:
        self.idle_timeout = 0
        # duplicated from clientconnection:
        self.notification_callbacks: dict[int, Callable] = {}
        self.send_notifications = False
        self.send_notifications_actions = False

    def init_from(self, _protocol, server) -> None:
        self.idle_timeout = server.idle_timeout
        self.session_name = server.session_name

    def init_state(self) -> None:
        self.last_user_event = monotonic()
        # grace duration is at least 10 seconds:
        self.idle_grace_duration = max(10, int(self.idle_timeout * (100 - GRACE_PERCENT) // 100))
        self.idle = False
        self.idle_timer = 0
        self.idle_grace_timer = 0

    def cleanup(self) -> None:
        self.cancel_idle_grace_timeout()
        self.cancel_idle_timeout()

    def get_info(self) -> dict[str, Any]:
        return {
            "idle_time": int(monotonic() - self.last_user_event),
            "idle": self.idle,
        }

    def parse_client_caps(self, _c: typedict) -> None:
        # start the timer
        self.schedule_idle_grace_timeout()
        self.schedule_idle_timeout()

    def user_event(self) -> None:
        log("user_event()")
        self.last_user_event = monotonic()
        self.cancel_idle_grace_timeout()
        self.schedule_idle_grace_timeout()
        self.cancel_idle_timeout()
        self.schedule_idle_timeout()
        if self.idle:
            self.no_idle()
        if self.notification_callbacks.pop(NotificationID.IDLE, None):
            self.notify_close(NotificationID.IDLE)

    def cancel_idle_timeout(self) -> None:
        it = self.idle_timer
        if it:
            self.idle_timer = 0
            GLib.source_remove(it)

    def schedule_idle_timeout(self) -> None:
        log("schedule_idle_timeout() idle_timer=%s, idle_timeout=%s", self.idle_timer, self.idle_timeout)
        if self.idle_timeout > 0:
            self.idle_timer = GLib.timeout_add(self.idle_timeout * 1000, self.idle_timedout)

    def cancel_idle_grace_timeout(self) -> None:
        igt = self.idle_grace_timer
        if igt:
            self.idle_grace_timer = 0
            GLib.source_remove(igt)

    def schedule_idle_grace_timeout(self) -> None:
        log("schedule_idle_grace_timeout() grace timer=%s, idle_timeout=%s", self.idle_grace_timer, self.idle_timeout)
        if self.idle_timeout > 0 and not self.is_closed():
            grace = self.idle_timeout - self.idle_grace_duration
            self.idle_grace_timer = GLib.timeout_add(max(0, int(grace * 1000)), self.idle_grace_timedout)
            log("schedule_idle_grace_timeout() timer=%s due in %i seconds", self.idle_grace_timer, grace)

    def idle_grace_timedout(self) -> None:
        self.idle_grace_timer = 0
        log("idle_grace_timedout()")
        if not self.send_notifications:
            # not much we can do!
            return
        # notify the user, giving him a chance to cancel the timeout:
        nid = NotificationID.IDLE
        if nid in self.notification_callbacks:
            return
        actions: Sequence[str] = ()
        if self.send_notifications_actions:
            actions = ("cancel", "Cancel Timeout")
        if self.session_name != "Xpra":
            summary = f"The Xpra session {self.session_name!r}"
        else:
            summary = "Xpra session"
        summary += " is about to timeout"
        body = "Unless this session sees some activity,\n" + \
               "it will be terminated soon."
        self.may_notify(nid, summary, body,
                        actions, {}, expire_timeout=10 * 1000,
                        icon_name="timer", user_callback=self.idle_notification_action)
        self.go_idle()

    def idle_notification_action(self, nid: int, action_id) -> None:
        log("idle_notification_action(%i, %s)", nid, action_id)
        if action_id == "cancel":
            self.user_event()

    def idle_timedout(self) -> None:
        self.idle_timer = 0
        p = self.protocol
        log("idle_timedout() protocol=%s", p)
        if p:
            self.disconnect(ConnectionMessage.IDLE_TIMEOUT)
        if not self.is_closed():
            self.schedule_idle_timeout()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.log import Logger
log = Logger("win32")

from libc.stdint cimport uintptr_t  # pylint: disable=syntax-error

cdef extern from "windows.h":
    ctypedef void *PVOID
    ctypedef PVOID HANDLE
    ctypedef HANDLE HWND
    ctypedef unsigned int UINT
    ctypedef unsigned long DWORD

    ctypedef Py_UNICODE WCHAR
    ctypedef const WCHAR* LPCWSTR
    ctypedef WCHAR* LPWSTR
    ctypedef LPWSTR LPTSTR
    ctypedef const char* LPCSTR

    int MultiByteToWideChar(UINT CodePage, DWORD dwFlags, LPCSTR lpMultiByteStr, int cbMultiByte, LPWSTR lpWideCharStr, int cchWideChar)
    int CP_ACP

cdef extern from "setappid.h" namespace "utility":
    int SetAppID(HWND hWnd, LPCWSTR pszAppID)

def set_window_group(hwnd, value):
    log("propsys.set_window_group(%i, %s)", hwnd, value)
    cdef HWND hWnd = <HWND> (<uintptr_t> hwnd)
    s = str(value).encode()
    cdef char *cstr = s
    cdef WCHAR[128] wstr
    cdef int r = MultiByteToWideChar(CP_ACP, 0, cstr, -1, wstr, 128)
    if r==0:
        log.warn("Warning: failed to convert string '%s' to wide win32 characters", s)
        log.warn(" MultiByteToWideChar returned %i", r)
        return r
    r = SetAppID(hWnd, wstr)
    log("propsys: SetAppID(%s, %s)=%i", <uintptr_t> hWnd, s, <uintptr_t> (&wstr))
    return r



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys


def main(args):
    from xpra.x11.bindings.posix_display_source import init_posix_display_source  # @UnresolvedImport
    init_posix_display_source()
    from xpra.x11.bindings.res import ResBindings  # @UnresolvedImport
    res = ResBindings()
    for x in args[1:]:
        try:
            if x.startswith("0x"):
                w = int(x[2:], 16)
            else:
                w = int(x)
        except Exception:
            print("cannot parse window number: %r" % x)
            continue
        pid = res.get_pid(w)
        cmdline = ""
        if pid:
            from xpra.util.io import load_binary_file
            cmdline = (load_binary_file("/proc/%i/cmdline" % pid) or b"").decode()
        print("pid(%#x)=%s      %s" % (w, pid, cmdline))

if __name__ == "__main__":
    main(sys.argv)



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import base64
import binascii
from collections.abc import Sequence

from xpra.util.str_fn import obsc
from xpra.util.env import envbool
from xpra.net.digest import get_salt
from xpra.auth.sys_auth_base import SysAuthenticator, log
from xpra.log import enable_color

SHOW = envbool("XPRA_OTP_SHOW", False)
SHOW_URI = envbool("XPRA_OTP_SHOW_URI", False)


def b32(s):
    try:
        assert base64.b32decode(s, casefold=True)
        return s
    except binascii.Error:
        return base64.b32encode(s.encode())


class Authenticator(SysAuthenticator):
    # DEFAULT_PROMPT = "OTP for '{username}'"
    DEFAULT_PROMPT = "OTP"

    def __init__(self, **kwargs):
        log("otp.Authenticator(%s)", kwargs)
        self.uid = -1
        self.gid = -1
        import pyotp  # @UnresolvedImport
        assert pyotp
        self.issuer_name = kwargs.pop("issuer-name", "Xpra")
        self.secret = b32(kwargs.pop("secret", pyotp.random_hex()))
        self.valid_window = int(kwargs.pop("valid-window", 0))
        # validate the base32 secret early:
        base64.b32decode(self.secret, casefold=True)
        super().__init__(**kwargs)
        if SHOW_URI:
            totp_uri = pyotp.totp.TOTP(self.secret).provisioning_uri(self.username, issuer_name=self.issuer_name)
            log("provisioning_uri=%s", totp_uri)

    def get_uid(self) -> int:
        return self.uid

    def get_gid(self) -> int:
        return self.gid

    def requires_challenge(self) -> bool:
        return True

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        if self.salt is not None:
            log.error("Error: authentication challenge already sent!")
            return b"", ""
        if "xor" not in digests:
            log.error("Error: xor is not supported by the client")
            return b"", ""
        self.salt = get_salt()
        self.digest = "xor"
        self.challenge_sent = True
        if SHOW:
            import pyotp  # @UnresolvedImport
            totp = pyotp.TOTP(self.secret)
            now = totp.now()
            log("otp current value for secret %s: %s", self.secret, now)
            log("recheck: %s", totp.verify(now))
        return self.salt, self.digest

    def check_password(self, password: str) -> bool:
        log("otp.check_password(%s)", obsc(password))
        import pyotp  # @UnresolvedImport
        totp = pyotp.TOTP(self.secret)
        r = totp.verify(password, valid_window=self.valid_window)
        log("otp.check_password(%s)=%s", obsc(password), r)
        if not r:
            raise ValueError("invalid OTP value")
        return True

    def __repr__(self):
        return "otp"


def main(argv) -> int:
    if len(argv) < 2 or len(argv) > 4:
        print(f"usage: {argv[0]} SECRET [username] [issuer-name]")
        return 1
    enable_color()
    import os
    secret = b32(argv[1])
    username = os.environ.get("USERNAME")
    issuer_name = "Xpra"
    if len(argv) >= 3:
        username = argv[2]
    if len(argv) >= 4:
        issuer_name = argv[3]
    import pyotp  # @UnresolvedImport
    totp_uri = pyotp.totp.TOTP(secret).provisioning_uri(username, issuer_name)
    log.info("provisioning_uri=%s", totp_uri)
    # qrcode module has problems - don't use it for now
    try:
        from xpra.gtk.dialogs.qrcode import show_qr
    except ImportError as e:
        log.info(" unable to show qr code: %s", e)
    else:
        show_qr(totp_uri)
    return 0


if __name__ == "__main__":
    import sys

    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import struct
from io import BytesIO
import PIL
from PIL import Image
from typing import Any
from collections.abc import Callable, Sequence

from xpra.common import SizedBuffer
from xpra.util.objects import typedict
from xpra.util.str_fn import csv, strtobytes, hexstr, memoryview_to_bytes
from xpra.codecs.debug import may_save_image
from xpra.log import Logger

log = Logger("encoder", "pillow")

Image.init()

DECODE_FORMATS = os.environ.get("XPRA_PILLOW_DECODE_FORMATS", "png,png/L,png/P,jpeg,webp").split(",")

PNG_HEADER = struct.pack("BBBBBBBB", 137, 80, 78, 71, 13, 10, 26, 10)


def is_png(data: bytes) -> bool:
    return data.startswith(PNG_HEADER)


RIFF_HEADER = b"RIFF"
WEBP_HEADER = b"WEBP"


def is_webp(data: bytes) -> bool:
    return data[:4] == RIFF_HEADER and data[8:12] == WEBP_HEADER


JPEG_HEADER = struct.pack("BBB", 0xFF, 0xD8, 0xFF)


def is_jpeg(data: bytes) -> bool:
    # the jpeg header is actually more complicated than this,
    # but in practice all the data we receive from the server
    # will have this type of header
    return data[:3] == JPEG_HEADER


def is_svg(data: bytes) -> bool:
    return strtobytes(data[:5]) == b"<?xml" or strtobytes(data[:4]) == b"<svg"


XPM_HEADER = b"/* XPM */"


def is_xpm(data: bytes) -> bool:
    return data[:9] == XPM_HEADER


def is_tiff(data: bytes) -> bool:
    if data[:2] == b"II":
        return data[2] == 42 and data[3] == 0
    if data[:2] == b"MM":
        return data[2] == 0 and data[3] == 42
    return False


HEADERS: dict[Callable[[bytes], bool], str] = {
    is_png: "png",
    is_webp: "webp",
    is_jpeg: "jpeg",
    is_svg: "svg",
    is_xpm: "xpm",
    is_tiff: "tiff",
}


def get_image_type(data: SizedBuffer) -> str:
    if not data:
        return ""
    if len(data) < 32:
        return ""
    header = memoryview_to_bytes(data[:32])
    for fn, encoding in HEADERS.items():
        if fn(header):
            return encoding
    return ""


def open_only(data: SizedBuffer, types=("png", "jpeg", "webp", "xpm")) -> Image:
    itype = get_image_type(data) or "unknown"
    if itype not in types:
        raise ValueError(f"invalid data: {itype}, not recognized as {csv(types)}, header: " + hexstr(data[:64]))
    buf = BytesIO(data)
    return Image.open(buf)


def get_version() -> str:
    return PIL.__version__


def get_type() -> str:
    return "pillow"


def do_get_encodings() -> list[str]:
    log("PIL.Image.OPEN=%s", Image.OPEN)
    encodings = []
    for encoding in DECODE_FORMATS:
        # strip suffix (so "png/L" -> "png")
        stripped = encoding.split("/")[0].upper()
        if stripped in Image.OPEN:
            encodings.append(encoding)
    log("do_get_encodings()=%s", encodings)
    return encodings


ENCODINGS: Sequence[str] = tuple(do_get_encodings())


def get_encodings() -> Sequence[str]:
    return ENCODINGS


def get_info() -> dict[str, Any]:
    return {
        "version": get_version(),
        "encodings": get_encodings(),
    }


def decompress(coding: str, img_data: bytes, options: typedict) -> tuple[str, bytes, int, int, int]:
    # can be called from any thread
    actual = get_image_type(img_data)
    if not actual or not coding.startswith(actual):
        raise ValueError(f"expected {coding!r} image data but received %r" % (actual or "unknown"))
    buf = BytesIO(img_data)
    img = Image.open(buf)
    assert img.mode in ("L", "LA", "P", "RGB", "RGBA", "RGBX"), f"invalid image mode: {img.mode}"
    transparency = options.intget("transparency", -1)
    if img.mode == "P":
        if transparency >= 0:
            # this deals with alpha without any extra work
            img = img.convert("RGBA")
        else:
            img = img.convert("RGB")
    elif img.mode == "L":
        if transparency >= 0:
            # why do we have to deal with alpha ourselves??
            def mask_value(a: int) -> int:
                if a != transparency:
                    return 255
                return 0

            mask = Image.eval(img, mask_value)
            mask = mask.convert("L")

            def nomask_value(a: int) -> int:
                if a != transparency:
                    return a
                return 0

            img = Image.eval(img, nomask_value)
            img = img.convert("RGBA")
            img.putalpha(mask)
        else:
            img = img.convert("RGB")
    elif img.mode == "LA":
        img = img.convert("RGBA")

    width, height = img.size
    if img.mode == "RGB":
        # PIL flattens the data to a continuous straightforward RGB format:
        rowstride = width * 3
        rgb_format = options.strget("rgb_format")
        rgb_format = rgb_format.replace("A", "").replace("X", "")
        # the webp encoder only takes BGRX input,
        # so we have to swap things around if it was fed "RGB":
        if rgb_format == "RGB":
            rgb_format = "BGR"
        else:
            rgb_format = "RGB"
    elif img.mode in ("RGBA", "RGBX"):
        rowstride = width * 4
        rgb_format = options.strget("rgb_format", img.mode)
        if coding == "webp":
            # the webp encoder only takes BGRX input,
            # so we have to swap things around if it was fed "RGBA":
            if rgb_format == "RGBA":
                rgb_format = "BGRA"
            elif rgb_format == "RGBX":
                rgb_format = "BGRX"
            elif rgb_format == "BGRA":
                rgb_format = "RGBA"
            elif rgb_format == "BGRX":
                rgb_format = "RGBX"
            else:
                log.warn("Warning: unexpected RGB format '%s'", rgb_format)
    else:
        raise ValueError(f"invalid image mode: {img.mode}")
    raw_data = img.tobytes("raw", img.mode)
    log("pillow decoded %7i bytes of %5s data to %8i bytes of %s", len(img_data), coding, len(raw_data), rgb_format)
    may_save_image(coding, img_data)
    return rgb_format, raw_data, width, height, rowstride


def selftest(_full=False) -> None:
    global ENCODINGS
    from xpra.codecs.checks import TEST_PICTURES  # pylint: disable=import-outside-toplevel
    # test data generated using the encoder:
    for encoding, test_data in TEST_PICTURES.items():
        if encoding not in ENCODINGS:
            # removed already
            continue
        for size, samples in test_data.items():
            log(f"testing {encoding} at size {size} with {len(samples)} samples")
            for i, (cdata, options) in enumerate(samples):
                try:
                    log(f"testing sample {i}: {len(cdata):5} bytes")
                    buf = BytesIO(cdata)
                    img = PIL.Image.open(buf)
                    assert img, "failed to open image data"
                    raw_data = img.tobytes("raw", img.mode)
                    assert raw_data
                    # now try with junk:
                    cdata = b"ABCD" + cdata
                    buf = BytesIO(cdata)
                    try:
                        img = PIL.Image.open(buf)
                        log.warn(f"Pillow failed to generate an error parsing invalid input: {img}, {options=}")
                    except Exception as e:
                        log("correctly raised exception for invalid input: %s", e)
                except Exception as e:
                    log("selftest:", exc_info=True)
                    log.error("Pillow error decoding %s with data:", encoding)
                    log.error(" %r", cdata)
                    log.error(" %s", e, exc_info=True)
                    ENCODINGS = tuple(x for x in ENCODINGS if x != encoding)


if __name__ == "__main__":
    selftest(True)
    print(csv(get_encodings()))



#!/usr/bin/env python3

import sys
import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
gi.require_version('Gdk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk     # pylint: disable=wrong-import-position @UnresolvedImport

width = 400
height = 200

def main():
    x, y = 0, 0
    if len(sys.argv)==3:
        x, y = int(sys.argv[1]), int(sys.argv[2])
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_size_request(width, height)
    window.connect("delete_event", Gtk.main_quit)
    window.realize()
    window.move(x, y)
    from xpra.x11.gtk.display_source import init_gdk_display_source
    init_gdk_display_source()
    from xpra.x11.bindings.window import X11WindowBindings  # @UnresolvedImport
    hints = {"position" : (x, y)}
    X11WindowBindings().setSizeHints(window.get_window().get_xid(), hints)
    window.show_all()
    Gtk.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from typing import Any
from ctypes.wintypes import HDC
from ctypes import (
    WinDLL,  # @UnresolvedImport
    c_void_p, Structure, c_int, c_uint, c_ulong, c_char, c_char_p, cast, pointer, POINTER,
)

from xpra.util.str_fn import Ellipsizer, strtobytes
from xpra.platform.win32.common import GetDeviceCaps
from xpra.platform.win32 import win32con
from xpra.platform.win32.ctypes_printing import GDIPrinterContext, DOCINFO, StartDocA, EndDoc, StartPage, EndPage, LPCSTR

LIBPDFIUMDLL = os.environ.get("XPRA_LIBPDFIUMDLL", "pdfium.dll")
try:
    pdfium = WinDLL(LIBPDFIUMDLL, use_last_error=True)
except OSError as e:  # @UndefinedVariable
    raise ImportError(f"cannot load {LIBPDFIUMDLL!r}: {e}") from None


class FPDF_LIBRARY_CONFIG(Structure):
    _fields_ = [
        ("version", c_int),
        ("m_pUserFontPaths", POINTER(POINTER(c_char))),
        ("m_pIsolate", c_void_p),
        ("m_v8EmbedderSlot", c_uint),
    ]


FPDF_DOCUMENT = c_void_p
FPDF_PAGE = c_void_p

FPDF_DestroyLibrary = pdfium.FPDF_DestroyLibrary
FPDF_InitLibraryWithConfig = pdfium.FPDF_InitLibraryWithConfig
FPDF_InitLibraryWithConfig.argtypes = [POINTER(FPDF_LIBRARY_CONFIG)]
FPDF_GetLastError = pdfium.FPDF_GetLastError
FPDF_GetLastError.restype = c_ulong
FPDF_GetPageCount = pdfium.FPDF_GetPageCount
FPDF_GetPageCount.argtypes = [FPDF_DOCUMENT]
FPDF_GetPageCount.restype = c_int
FPDF_LoadPage = pdfium.FPDF_LoadPage
FPDF_LoadPage.argtypes = [FPDF_DOCUMENT, c_int]
FPDF_LoadPage.restype = FPDF_PAGE
FPDF_RenderPage = pdfium.FPDF_RenderPage
FPDF_RenderPage.argtypes = [HDC, FPDF_PAGE, c_int, c_int, c_int, c_int, c_int, c_int]
FPDF_LoadMemDocument = pdfium.FPDF_LoadMemDocument
FPDF_LoadMemDocument.restype = FPDF_DOCUMENT
FPDF_LoadMemDocument.argtypes = [c_void_p, c_int, c_void_p]
FPDF_CloseDocument = pdfium.FPDF_CloseDocument
FPDF_CloseDocument.argtypes = [FPDF_DOCUMENT]
FPDF_ClosePage = pdfium.FPDF_ClosePage
FPDF_ClosePage.argtypes = [FPDF_PAGE]

FPDF_ERR_SUCCESS = 0  # No error.
FPDF_ERR_UNKNOWN = 1  # Unknown error.
FPDF_ERR_FILE = 2  # File not found or could not be opened.
FPDF_ERR_FORMAT = 3  # File not in PDF format or corrupted.
FPDF_ERR_PASSWORD = 4  # Password required or incorrect password.
FPDF_ERR_SECURITY = 5  # Unsupported security scheme.
FPDF_ERR_PAGE = 6  # Page not found or content error.
FPDF_ERR_XFALOAD = 7  # Load XFA error.
FPDF_ERR_XFALAYOUT = 8  # Layout XFA error.

ERROR_STR = {
    # FPDF_ERR_SUCCESS : No error.
    FPDF_ERR_UNKNOWN: "Unknown error",
    FPDF_ERR_FILE: "File not found or could not be opened",
    FPDF_ERR_FORMAT: "File not in PDF format or corrupted",
    FPDF_ERR_PASSWORD: "Password required or incorrect password",
    FPDF_ERR_SECURITY: "Unsupported security scheme",
    FPDF_ERR_PAGE: "Page not found or content error",
    FPDF_ERR_XFALOAD: "Load XFA error",
    FPDF_ERR_XFALAYOUT: "Layout XFA error",
}

FPDF_ANNOT = 0x01
FPDF_LCD_TEXT = 0x02
FPDF_NO_NATIVETEXT = 0x04
FPDF_GRAYSCALE = 0x08
FPDF_DEBUG_INFO = 0x80
FPDF_NO_CATCH = 0x100
FPDF_RENDER_LIMITEDIMAGECACHE = 0x200
FPDF_RENDER_FORCEHALFTONE = 0x400
FPDF_PRINTING = 0x800
FPDF_RENDER_NO_SMOOTHTEXT = 0x1000
FPDF_RENDER_NO_SMOOTHIMAGE = 0x2000
FPDF_RENDER_NO_SMOOTHPATH = 0x4000
FPDF_REVERSE_BYTE_ORDER = 0x10


def get_error():
    v = FPDF_GetLastError()
    return ERROR_STR.get(v, v)


def do_print_pdf(hdc, title=b"PDF Print Test", pdf_data=None):
    assert pdf_data, "no pdf data"
    from xpra.log import Logger
    log = Logger("printing", "win32")
    log("pdfium=%s", pdfium)
    buf = c_char_p(pdf_data)
    log("pdf data buffer: %s", Ellipsizer(pdf_data))
    log("FPDF_InitLibraryWithConfig=%s", FPDF_InitLibraryWithConfig)
    config = FPDF_LIBRARY_CONFIG()
    config.m_pUserFontPaths = None
    config.version = 2
    config.m_pIsolate = None
    config.m_v8EmbedderSlot = 0
    FPDF_InitLibraryWithConfig(config)
    x = 0
    y = 0
    w = GetDeviceCaps(hdc, win32con.HORZRES)
    h = GetDeviceCaps(hdc, win32con.VERTRES)
    rotate = 0
    log("printer device size: %ix%i", w, h)
    flags = FPDF_PRINTING | FPDF_DEBUG_INFO
    try:
        doc = FPDF_LoadMemDocument(cast(buf, c_void_p), len(pdf_data), None)
        if not doc:
            log.error("Error: FPDF_LoadMemDocument failed, error: %s", get_error())
            return -1
        log("FPDF_LoadMemDocument(..)=%s", doc)
        count = FPDF_GetPageCount(doc)
        log("FPDF_GetPageCount(%s)=%s", doc, count)
        docinfo = DOCINFO()
        docinfo.lpszDocName = LPCSTR(b"%s\0" % title)
        jobid = StartDocA(hdc, pointer(docinfo))
        if jobid < 0:
            log.error("Error: StartDocA failed: %i", jobid)
            return jobid
        log("StartDocA()=%i", jobid)
        try:
            for i in range(count):
                page = FPDF_LoadPage(doc, i)
                if not page:
                    log.error("Error: FPDF_LoadPage failed for page %i, error: %s", i, get_error())
                    return -2
                log("FPDF_LoadPage()=%s page %i loaded", page, i)
                StartPage(hdc)
                FPDF_RenderPage(hdc, page, x, y, w, h, rotate, flags)
                EndPage(hdc)
                FPDF_ClosePage(page)
                log("FPDF_RenderPage page %i rendered", i)
        finally:
            FPDF_CloseDocument(doc)
            EndDoc(hdc)
    finally:
        FPDF_DestroyLibrary()
    return jobid


def print_pdf(printer_name, title, pdf_data):
    with GDIPrinterContext(printer_name) as hdc:
        return do_print_pdf(hdc, title, pdf_data)


EXIT = False
JOBS_INFO: dict[int, dict[str, Any]] = {}


def watch_print_job_status():
    from xpra.log import Logger
    log = Logger("printing", "win32")
    log("wait_for_print_job_end()")
    # log("wait_for_print_job_end(%i)", print_job_id)
    from xpra.platform.win32.printer_notify import wait_for_print_job_info, job_status
    while not EXIT:
        info = wait_for_print_job_info(timeout=1.0)
        if not info:
            continue
        log("wait_for_print_job_info()=%s", info)
        for nd in info:
            job_id, key, value = nd
            if key == "job_status":
                value = job_status(value)
            log("job_id=%s, key=%s, value=%s", job_id, key, value)
            JOBS_INFO.setdefault(int(job_id), {})[key] = value


def main():
    # pylint: disable=import-outside-toplevel
    if len(sys.argv) == 2:
        from xpra.platform.win32.printing import get_printers
        printers = get_printers()
        printer_name = strtobytes(printers.keys()[0])
    elif len(sys.argv) in (3, 4):
        printer_name = strtobytes(sys.argv[2])
    else:
        print(f"usage: {sys.argv[0]} /path/to/document.pdf [printer-name] [document-title]")
        return -3

    filename = sys.argv[1]
    with open(filename, "rb") as f:
        pdf_data = f.read()

    if len(sys.argv) == 4:
        title = strtobytes(sys.argv[3])
    else:
        title = strtobytes(os.path.basename(filename))

    import time
    from xpra.util.str_fn import csv
    from xpra.log import Logger
    log = Logger("printing", "win32")

    # start a new thread before submitting the document,
    # because otherwise the job may complete before we can get its status
    from threading import Thread
    t = Thread(target=watch_print_job_status, name="watch print job status")
    t.daemon = True
    t.start()

    job_id = print_pdf(printer_name, title, pdf_data)
    if job_id < 0:
        return job_id
    # wait for job to end:
    job_status: list[str] = []
    while True:
        job_info = JOBS_INFO.get(job_id, {})
        log("job_info[%i]=%s", job_id, job_info)
        v = job_info.get("job_status")
        if v != job_status:
            log.info("print job status: %s", csv(v))
            job_status = v
            if "OFFLINE" in job_status or "DELETING" in job_status:
                global EXIT
                EXIT = True
                break
        time.sleep(1.0)
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from time import monotonic
from collections import deque
from threading import Lock
from typing import Any, Literal, NoReturn

from xpra.audio.audio_pipeline import AudioPipeline
from xpra.gstreamer.common import (
    normv, make_buffer, plugin_str,
    get_default_appsrc_attributes, get_element_str,
    GST_FLOW_OK,
)
from xpra.audio.gstreamer_util import (
    get_decoder_elements, has_plugins,
    get_queue_time, get_decoders,
    get_default_sink_plugin, get_sink_plugins,
    MP3, CODEC_ORDER, QUEUE_LEAK,
    GST_QUEUE_NO_LEAK, MS_TO_NS, DEFAULT_SINK_PLUGIN_OPTIONS,
)
from xpra.gtk.gobject import one_arg_signal
from xpra.net.compression import decompress_by_name
from xpra.scripts.config import InitExit
from xpra.os_util import gi_import
from xpra.util.str_fn import csv
from xpra.util.env import envint, envbool
from xpra.util.thread import start_thread
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("audio")
gstlog = Logger("gstreamer")

GObject = gi_import("GObject")

SINK_SHARED_DEFAULT_ATTRIBUTES: dict[str, Any] = {
    "sync": False,
}
NON_AUTO_SINK_ATTRIBUTES: dict[str, Any] = {
    "async": True,
    "qos": True,
}

SINK_DEFAULT_ATTRIBUTES = {
    "pulsesink": {"client-name": "Xpra"},
}

QUEUE_SILENT = envbool("XPRA_QUEUE_SILENT", False)
QUEUE_TIME = get_queue_time(450)

UNMUTE_DELAY = envint("XPRA_UNMUTE_DELAY", 1000)
GRACE_PERIOD = envint("XPRA_SOUND_GRACE_PERIOD", 2000)
# percentage: from 0 for no margin, to 200% which triples the buffer target
MARGIN = max(0, min(200, envint("XPRA_SOUND_MARGIN", 50)))
# how high we push up the min-level to prevent underruns:
UNDERRUN_MIN_LEVEL = max(0, envint("XPRA_SOUND_UNDERRUN_MIN_LEVEL", 150))
CLOCK_SYNC = envbool("XPRA_CLOCK_SYNC", False)


def uncompress_data(data, metadata: dict):
    if not data or not metadata:
        return data
    compress = metadata.get("compress")
    if not compress:
        return data
    if compress != "lz4":
        raise ValueError(f"unsupported compresssion {compress!r}")
    v = decompress_by_name(data, compress)
    # log("decompressed %s data: %i bytes into %i bytes", compress, len(data), len(v))
    return v


class AudioSink(AudioPipeline):
    __gsignals__ = AudioPipeline.__generic_signals__.copy()
    __gsignals__ |= {
        "eos": one_arg_signal,
    }

    def __init__(self, sink_type=None, sink_options=None, codecs=(), codec_options=None, volume=1.0):
        if not sink_type:
            sink_type = get_default_sink_plugin()
        if sink_type not in get_sink_plugins():
            raise InitExit(1, "invalid sink: %s" % sink_type)
        matching = [x for x in CODEC_ORDER if (x in codecs and x in get_decoders())]
        log("AudioSink(..) found matching codecs %s", matching)
        if not matching:
            raise InitExit(1, "no matching codecs between arguments '%s' and supported list '%s'" % (
                csv(codecs), csv(get_decoders().keys())))
        codec = matching[0]
        decoder, parser, stream_compressor = get_decoder_elements(codec)
        super().__init__(codec)
        self.container_format = (parser or "").replace("demux", "").replace("depay", "")
        self.sink_type = sink_type
        self.stream_compressor = stream_compressor
        log("container format=%s, stream_compressor=%s, sink type=%s",
            self.container_format, self.stream_compressor, self.sink_type)
        self.levels = deque(maxlen=100)
        self.volume = None
        self.src = None
        self.sink = None
        self.queue = None
        self.normal_volume = volume
        self.target_volume = volume
        self.volume_timer = 0
        self.overruns = 0
        self.underruns = 0
        self.overrun_events = deque(maxlen=100)
        self.queue_state = "starting"
        self.last_data = None
        self.last_underrun = 0
        self.last_overrun = 0
        self.refill = True
        self.last_max_update = monotonic()
        self.last_min_update = monotonic()
        self.level_lock = Lock()
        pipeline_els = [get_element_str("appsrc", get_default_appsrc_attributes())]
        if parser:
            pipeline_els.append(parser)
        if decoder:
            decoder_str = plugin_str(decoder, codec_options)
            pipeline_els.append(decoder_str)
        pipeline_els.append("audioconvert")
        pipeline_els.append("audioresample")
        if QUEUE_TIME > 0:
            pipeline_els.append(get_element_str("queue", {
                "name": "queue",
                "min-threshold-time": 0,
                "max-size-buffers": 0,
                "max-size-bytes": 0,
                "max-size-time": QUEUE_TIME,
                "leaky": QUEUE_LEAK,
            }))
        pipeline_els.append(get_element_str("volume", {"name": "volume", "volume": 0}))
        if CLOCK_SYNC:
            if not has_plugins("clocksync"):
                log.warn("Warning: cannot enable clocksync, element not found")
            else:
                pipeline_els.append("clocksync")
        sink_attributes = SINK_SHARED_DEFAULT_ATTRIBUTES.copy()
        # anything older than this may cause problems (ie: centos 6.x)
        # because the attributes may not exist
        sink_attributes.update(SINK_DEFAULT_ATTRIBUTES.get(sink_type, {}))
        get_options_cb = DEFAULT_SINK_PLUGIN_OPTIONS.get(sink_type.replace("sink", ""))
        if get_options_cb:
            v = get_options_cb()
            log("%s()=%s", get_options_cb, v)
            sink_attributes.update(v)
        if sink_options:
            sink_attributes.update(sink_options)
        sink_attributes["name"] = "sink"
        if sink_type != "autoaudiosink":
            sink_attributes.update(NON_AUTO_SINK_ATTRIBUTES)
        sink_str = plugin_str(sink_type, sink_attributes)
        pipeline_els.append(sink_str)
        if not self.setup_pipeline_and_bus(pipeline_els):
            return
        self.volume = self.pipeline.get_by_name("volume")
        self.src = self.pipeline.get_by_name("src")
        self.sink = self.pipeline.get_by_name("sink")
        self.queue = self.pipeline.get_by_name("queue")
        if self.queue:
            if QUEUE_SILENT:
                self.queue.set_property("silent", False)
            else:
                self.queue.connect("overrun", self.queue_overrun)
                self.queue.connect("underrun", self.queue_underrun)
                self.queue.connect("running", self.queue_running)
                self.queue.connect("pushing", self.queue_pushing)
        self.init_file(codec)

    def __repr__(self):  # pylint: disable=arguments-differ
        return "AudioSink('%s' - %s)" % (self.pipeline_str, self.state)

    def cleanup(self) -> None:
        super().cleanup()
        self.cancel_volume_timer()
        self.sink_type = ""
        self.src = None

    def start(self) -> bool:
        if not super().start():
            return False
        GLib.timeout_add(UNMUTE_DELAY, self.start_adjust_volume)
        return True

    def start_adjust_volume(self, interval: int = 100) -> bool:
        if self.volume_timer != 0:
            GLib.source_remove(self.volume_timer)
        self.volume_timer = GLib.timeout_add(interval, self.adjust_volume)
        return False

    def cancel_volume_timer(self) -> None:
        if self.volume_timer != 0:
            GLib.source_remove(self.volume_timer)
            self.volume_timer = 0

    def adjust_volume(self) -> bool:
        if not self.volume:
            self.volume_timer = 0
            return False
        cv = self.volume.get_property("volume")
        delta = self.target_volume - cv
        from math import sqrt, copysign
        change = copysign(sqrt(abs(delta)), delta) / 15.0
        gstlog("adjust_volume current volume=%.2f, change=%.2f", cv, change)
        self.volume.set_property("volume", max(0, cv + change))
        if abs(delta) < 0.01:
            self.volume_timer = 0
            return False
        return True

    def queue_pushing(self, *_args) -> Literal[True]:
        gstlog("queue_pushing")
        self.queue_state = "pushing"
        self.emit_info()
        return True

    def queue_running(self, *_args) -> Literal[True]:
        gstlog("queue_running")
        self.queue_state = "running"
        self.emit_info()
        return True

    def queue_underrun(self, *_args) -> Literal[True]:
        now = monotonic()
        if self.queue_state == "starting" or 1000 * (now - self.start_time) < GRACE_PERIOD:
            gstlog("ignoring underrun during startup")
            return True
        self.underruns += 1
        gstlog("queue_underrun")
        self.queue_state = "underrun"
        if now - self.last_underrun > 5:
            # only count underruns when we're back to no min time:
            qmin = self.queue.get_property("min-threshold-time") // MS_TO_NS
            clt = self.queue.get_property("current-level-time") // MS_TO_NS
            gstlog("queue_underrun level=%3i, min=%3i", clt, qmin)
            if qmin == 0 and clt < 10:
                self.last_underrun = now
                self.refill = True
                self.set_max_level()
                self.set_min_level()
        self.emit_info()
        return True

    def get_level_range(self, mintime=2, maxtime=10) -> int:
        now = monotonic()
        filtered = [v for t, v in tuple(self.levels) if mintime <= (now - t) <= maxtime]
        if len(filtered) >= 10:
            maxl = max(filtered)
            minl = min(filtered)
            # range of the levels recorded:
            return maxl - minl
        return 0

    def queue_overrun(self, *_args) -> Literal[True]:
        now = monotonic()
        if self.queue_state == "starting" or 1000 * (now - self.start_time) < GRACE_PERIOD:
            gstlog("ignoring overrun during startup")
            return True
        clt = self.queue.get_property("current-level-time") // MS_TO_NS
        log("queue_overrun level=%ims", clt)
        now = monotonic()
        # grace period of recording overruns:
        # (because when we record an overrun, we lower the max-time,
        # which causes more overruns!)
        if now - self.last_overrun > 2:
            self.last_overrun = now
            self.set_max_level()
            self.overrun_events.append(now)
        self.overruns += 1
        return True

    def set_min_level(self) -> None:
        if not self.queue:
            return
        now = monotonic()
        elapsed = now - self.last_min_update
        lrange = self.get_level_range()
        log("set_min_level() lrange=%i, elapsed=%i", lrange, elapsed)
        if elapsed < 1:
            # not more than once a second
            return
        if self.refill:
            # need to have a gap between min and max,
            # so we cannot go higher than mst-50:
            mst = self.queue.get_property("max-size-time") // MS_TO_NS
            mrange = max(lrange + 100, UNDERRUN_MIN_LEVEL)
            mtt = min(mst - 50, mrange)
            gstlog("set_min_level mtt=%3i, max-size-time=%3i, lrange=%s, mrange=%s (UNDERRUN_MIN_LEVEL=%s)",
                   mtt, mst, lrange, mrange, UNDERRUN_MIN_LEVEL)
        else:
            mtt = 0
        cmtt = self.queue.get_property("min-threshold-time") // MS_TO_NS
        if cmtt == mtt:
            return
        if not self.level_lock.acquire(False):
            gstlog("cannot get level lock for setting min-threshold-time")
            return
        try:
            self.queue.set_property("min-threshold-time", mtt * MS_TO_NS)
            gstlog("set_min_level min-threshold-time=%s", mtt)
            self.last_min_update = now
        finally:
            self.level_lock.release()

    def set_max_level(self) -> None:
        if not self.queue:
            return
        now = monotonic()
        elapsed = now - self.last_max_update
        if elapsed < 1:
            # not more than once a second
            return
        lrange = self.get_level_range(mintime=0)
        log("set_max_level lrange=%3i, elapsed=%is", lrange, int(elapsed))
        cmst = self.queue.get_property("max-size-time") // MS_TO_NS
        # overruns in the last minute:
        olm = len([x for x in tuple(self.overrun_events) if now - x < 60])
        # increase target if we have more than 5 overruns in the last minute:
        target_mst = lrange * (100 + MARGIN + min(100, olm * 20)) // 100
        # from 100% down to 0% in 2 seconds after underrun:
        pct = max(0, int((self.last_overrun + 2 - now) * 50))
        # use this last_overrun percentage value to temporarily decrease the target
        # (causes overruns that drop packets and lower the buffer level)
        target_mst = max(50, int(target_mst - pct * lrange // 100))
        mst = (cmst + target_mst) // 2
        if self.refill:
            # temporarily raise max level during underruns,
            # so set_min_level has more room for manoeuver:
            mst += UNDERRUN_MIN_LEVEL
        # cap it at 1 second:
        mst = min(mst, 1000)
        log("set_max_level overrun count=%-2i, margin=%3i, pct=%2i, cmst=%3i, target=%3i, mst=%3i",
            olm, MARGIN, pct, cmst, target_mst, mst)
        if abs(cmst - mst) <= max(50, lrange // 2):
            # not enough difference
            return
        if not self.level_lock.acquire(False):
            gstlog("cannot get level lock for setting max-size-time")
            return
        try:
            self.queue.set_property("max-size-time", mst * MS_TO_NS)
            log("set_max_level max-size-time=%s", mst)
            self.last_max_update = now
        finally:
            self.level_lock.release()

    def eos(self) -> int:
        gstlog("eos()")
        if self.src:
            self.src.emit('end-of-stream')
        self.cleanup()
        return GST_FLOW_OK

    def get_info(self) -> dict[str, Any]:
        info = super().get_info()
        if QUEUE_TIME > 0 and self.queue:
            clt = self.queue.get_property("current-level-time")
            qmax = self.queue.get_property("max-size-time")
            qmin = self.queue.get_property("min-threshold-time")
            info["queue"] = {
                "min": qmin // MS_TO_NS,
                "max": qmax // MS_TO_NS,
                "cur": clt // MS_TO_NS,
                "pct": min(QUEUE_TIME, clt) * 100 // qmax,
                "overruns": self.overruns,
                "underruns": self.underruns,
                "state": self.queue_state,
            }
        info["sink"] = self.get_element_properties(
            self.sink,
            "buffer-time", "latency-time",
            # "next_sample", "eos_rendering",
            "async", "blocksize",
            "enable-last-sample",
            "max-bitrate", "max-lateness",
            # "processing-deadline",
            "qos", "render-delay", "sync",
            "throttle-time", "ts-offset",
            ignore_missing=True
        )
        return info

    def can_push_buffer(self) -> bool:
        if not self.src:
            log("no source, dropping buffer")
            return False
        if self.state in ("stopped", "error"):
            log("pipeline is %s, dropping buffer", self.state)
            return False
        return True

    def add_data(self, data: bytes, metadata: dict, packet_metadata=()) -> None:
        if not self.can_push_buffer():
            return
        data = uncompress_data(data, metadata)
        for x in packet_metadata:
            self.do_add_data(x)
        if self.do_add_data(data, metadata):
            self.rec_queue_level(data)
            self.set_max_level()
            self.set_min_level()
            # drop back down quickly if the level has reached min:
            if self.refill:
                clt = self.queue.get_property("current-level-time") // MS_TO_NS
                qmin = self.queue.get_property("min-threshold-time") // MS_TO_NS
                gstlog("add_data: refill=%s, level=%i, min=%i", self.refill, clt, qmin)
                if 0 < qmin < clt:
                    self.refill = False
        self.emit_info()

    def do_add_data(self, data, metadata=None) -> bool:
        # having a timestamp causes problems with the queue and overruns:
        log("do_add_data(%s bytes, %s) queue_state=%s", len(data), metadata, self.queue_state)
        self.save_to_file(data)
        buf = make_buffer(data)
        if metadata:
            # having a timestamp causes problems with the queue and overruns:
            # ts = metadata.get("timestamp")
            # if ts is not None:
            #    buf.timestamp = normv(ts)
            #    log.info("timestamp=%s", ts)
            d = metadata.get("duration")
            if d is not None:
                d = normv(d)
                if d > 0:
                    buf.duration = normv(d)
        if self.push_buffer(buf) == GST_FLOW_OK:
            self.inc_buffer_count()
            self.inc_byte_count(len(data))
            return True
        return False

    def rec_queue_level(self, data) -> None:
        q = self.queue
        if not q:
            return
        clt = q.get_property("current-level-time") // MS_TO_NS
        log("pushed %5i bytes, new buffer level: %3ims, queue state=%s", len(data), clt, self.queue_state)
        now = monotonic()
        self.levels.append((now, clt))

    def push_buffer(self, buf) -> int:
        # buf.size = size
        # buf.timestamp = timestamp
        # buf.duration = duration
        # buf.offset = offset
        # buf.offset_end = offset_end
        # buf.set_caps(gst.caps_from_string(caps))
        r = self.src.emit("push-buffer", buf)
        if r == GST_FLOW_OK:
            return r
        if self.queue_state != "error":
            log.error("Error pushing buffer: %s", r)
            self.update_state("error")
            self.emit('error', "push-buffer error: %s" % r)
        return 1


GObject.type_register(AudioSink)


def main() -> int:
    from xpra.platform import program_context
    with program_context("Audio-Record"):
        args = sys.argv
        log.enable_debug()
        import os.path
        if len(args) not in (2, 3):
            print("usage: %s [-v|--verbose] filename [codec]" % sys.argv[0])
            return 1
        filename = args[1]
        if not os.path.exists(filename):
            print("file %s does not exist" % filename)
            return 2
        decoders = get_decoders()
        if len(args) == 3:
            codec = args[2]
            if codec not in decoders:
                print("invalid codec: %s" % codec)
                print("only supported: %s" % str(decoders.keys()))
                return 2
            codecs = [codec]
        else:
            codec = None
            parts = filename.split(".")
            if len(parts) > 1:
                extension = parts[-1]
                if extension.lower() in decoders:
                    codec = extension.lower()
                    print("guessed codec %s from file extension %s" % (codec, extension))
            if codec is None:
                print("assuming this is an mp3 file...")
                codec = MP3
            codecs = [codec]

        log.enable_debug()
        with open(filename, "rb") as f:
            data = f.read()
        print("loaded %s bytes from %s" % (len(data), filename))
        # force no leak since we push all the data at once
        from xpra.audio import gstreamer_util
        gstreamer_util.QUEUE_LEAK = GST_QUEUE_NO_LEAK
        gstreamer_util.QUEUE_SILENT = True
        ss = AudioSink(codecs=codecs)

        def eos(*args):
            print("eos%s" % (args,))
            GLib.idle_add(glib_mainloop.quit)

        ss.connect("eos", eos)
        ss.start()

        glib_mainloop = GLib.MainLoop()

        import signal

        def deadly_signal(*_args) -> None:
            GLib.idle_add(ss.stop)
            GLib.idle_add(glib_mainloop.quit)

            def force_quit(_sig, _frame) -> NoReturn:
                sys.exit()

            signal.signal(signal.SIGINT, force_quit)
            signal.signal(signal.SIGTERM, force_quit)

        signal.signal(signal.SIGINT, deadly_signal)
        signal.signal(signal.SIGTERM, deadly_signal)

        def check_for_end(*_args):
            qtime = ss.queue.get_property("current-level-time") // MS_TO_NS
            if qtime <= 0:
                log.info("underrun (end of stream)")
                start_thread(ss.stop, "stop", daemon=True)
                GLib.timeout_add(500, glib_mainloop.quit)
                return False
            return True

        GLib.timeout_add(1000, check_for_end)
        GLib.idle_add(ss.add_data, data)

        glib_mainloop.run()
        return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

import sys

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk     # pylint: disable=wrong-import-position @UnresolvedImport
from xpra.util.env import envbool


class StatusIcon:
    def __init__(self, name="test", tooltip="StatusIcon Example"):
        self.name = name
        self.statusicon = Gtk.StatusIcon()
        self.counter = 0
        self.statusicon.set_name(name)
        self.statusicon.set_from_stock(Gtk.STOCK_HOME)
        self.statusicon.connect("popup-menu", self.popup_menu)
        self.statusicon.connect("activate", self.activate)
        self.statusicon.set_tooltip_text(tooltip)
        #build list of stock icons:
        self.stock = {}
        try:
            nc = []
            if envbool("XPRA_NATIVE_NOTIFIER", True):
                from xpra.platform.gui import get_native_notifier_classes
                nc += get_native_notifier_classes()
            from xpra.gtk.notifier import GTKNotifier
            nc.append(GTKNotifier)
            self.notifier = nc[0](self.notification_closed, self.notification_action)
            self.notifier.app_name_format = "%s"
            #ensure we can send the image-path hint with the dbus backend:
            if hasattr(self.notifier, "noparse_hints"):
                self.notifier.parse_hints = self.notifier.noparse_hints
        except Exception as e:
            import traceback
            traceback.print_stack()
            print("Failed to instantiate the notifier: %s" % e)
        self.nid = 1
        for x in dir(Gtk):
            if x.startswith("STOCK_"):
                self.stock[x[len("STOCK_"):]] = getattr(Gtk, x)

    def activate(self, *_args):
        self.counter += 1
        name, stock = list(self.stock.items())[self.counter % len(self.stock)]
        print("setting tray icon to: %s" % name)
        self.statusicon.set_from_stock(stock)

    def popup_menu(self, icon, button, time):
        menu = Gtk.Menu()
        quit_menu = Gtk.MenuItem("Quit")
        quit_menu.connect("activate", Gtk.main_quit)
        menu.append(quit_menu)
        notify_menu = Gtk.MenuItem("Send Notification")
        notify_menu.connect("activate", self.notify)
        menu.append(notify_menu)
        menu.show_all()
        menu.popup(None, None, Gtk.StatusIcon.position_menu, self.statusicon, button, time)

    def notification_closed(self, nid, reason, text):
        print("notification_closed(%i, %i, %s)" % (nid, reason, text))

    def notification_action(self, nid, action):
        print("notification_action(%s, %s)" % (nid, action))

    def notify(self, *_args):
        actions = ["0", "Hello", "1", "Goodbye"]
        hints = {
            "image-path"    : "/usr/share/xpra/icons/encoding.png",
            }
        self.notifier.show_notify("dbus-id", None, self.nid, self.name, 0,
                                  "", "Notification Summary", "Notification Body", actions, hints, 60*1000, "")
        self.nid += 1


def main():
    name = "test"
    if len(sys.argv)>=2:
        name = sys.argv[1]
    tooltip = "StatusIcon Example"
    if len(sys.argv)>=3:
        tooltip = sys.argv[2]
    from xpra.platform import program_context
    with program_context(name, name):
        StatusIcon(name, tooltip)
        Gtk.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=import-outside-toplevel

from typing import Any
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass

from xpra.log import Logger
from xpra.common import SizedBuffer
from xpra.net.protocol.header import FLAGS_RENCODE, FLAGS_RENCODEPLUS, FLAGS_YAML, FLAGS_NOHEADER, pack_header
from xpra.util.str_fn import strtobytes, memoryview_to_bytes
from xpra.util.env import envbool

# all the encoders we know about:
ALL_ENCODERS: Sequence[str] = ("rencodeplus", "rencode", "bencode", "yaml", "none")

VALID_ENCODERS: Sequence[str] = ("rencodeplus", "yaml", "none")
# the encoders we may have, in the best compatibility order
TRY_ENCODERS: Sequence[str] = ("rencodeplus", "yaml", "none")
# order for performance:
PERFORMANCE_ORDER: Sequence[str] = ("rencodeplus", "yaml", )


@dataclass
class PacketEncoder:
    name: str
    flag: int
    version: str
    encode: Callable[[Any], tuple[SizedBuffer, int]]
    decode: Callable[[SizedBuffer], Any]


ENCODERS: dict[str, PacketEncoder] = {}


def init_rencodeplus() -> PacketEncoder:
    from xpra.net.rencodeplus import rencodeplus  # type: ignore[attr-defined]
    rencodeplus_dumps = rencodeplus.dumps

    def do_rencodeplus(value) -> tuple[SizedBuffer, int]:
        return rencodeplus_dumps(value), FLAGS_RENCODEPLUS

    return PacketEncoder("rencodeplus", FLAGS_RENCODEPLUS, rencodeplus.__version__, do_rencodeplus, rencodeplus.loads)


def init_yaml() -> PacketEncoder:
    import yaml

    def represent_tuple(dumper, value):
        return dumper.represent_list(value)

    yaml.add_representer(tuple, represent_tuple)

    def yaml_encode(value) -> tuple[SizedBuffer, int]:
        return yaml.dump(value).encode("utf-8"), FLAGS_YAML

    def yaml_decode(value) -> Any:
        return yaml.load(value.decode("utf-8"), Loader=yaml.SafeLoader)

    return PacketEncoder("yaml", FLAGS_YAML, yaml.__version__, yaml_encode, yaml_decode)


def b(value) -> bytes:
    if isinstance(value, Iterable) and not isinstance(value, (str, bytes, memoryview, bytearray)):
        return b" ".join(b(item) for item in value)
    return memoryview_to_bytes(value)


def none_encode(value) -> tuple[bytes, int]:
    # just send data as a byte string for clients that don't understand xpra packet format:
    bdata = b(value) + b"\n"
    return bdata, FLAGS_NOHEADER


def none_decode(data: SizedBuffer) -> SizedBuffer:
    return data


def init_none() -> PacketEncoder:
    return PacketEncoder("none", FLAGS_NOHEADER, "0", none_encode, none_decode)


def init_encoders(*names: str) -> None:
    for x in names:
        if x not in TRY_ENCODERS:
            logger = Logger("network", "protocol")
            logger.warn("Warning: invalid encoder '%s'", x)
            continue
        if not envbool("XPRA_" + x.upper(), True):
            continue
        fn = globals().get(f"init_{x}")
        try:
            assert callable(fn)
            e = fn()
            assert e
            ENCODERS[x] = e
        except (ImportError, AttributeError, AssertionError):
            logger = Logger("network", "protocol")
            logger.debug("no %s", x, exc_info=True)


def init_all() -> None:
    init_encoders(*(list(TRY_ENCODERS) + ["none"]))


def get_packet_encoding_caps(full_info: int = 1) -> dict[str, Any]:
    caps: dict[str, Any] = {}
    for name in TRY_ENCODERS:
        d = caps.setdefault(name, {})
        e = ENCODERS.get(name)
        d[""] = e is not None
        if e is None:
            continue
        if full_info > 1 and e.version:
            d["version"] = e.version
    return caps


def get_enabled_encoders(order: Iterable[str] = TRY_ENCODERS) -> Sequence[str]:
    return tuple(x for x in order if x in ENCODERS)


def get_encoder(e) -> Callable:
    if e not in VALID_ENCODERS:
        raise ValueError(f"invalid encoder name {e!r}")
    if e not in ENCODERS:
        raise ValueError(f"{e!r} is not available")
    return ENCODERS[e].encode


def get_packet_encoding_type(protocol_flags: int) -> str:
    if protocol_flags & FLAGS_RENCODEPLUS:
        return "rencodeplus"
    if protocol_flags & FLAGS_RENCODE:
        return "rencode"
    if protocol_flags & FLAGS_YAML:
        return "yaml"
    return "bencode"


class InvalidPacketEncodingException(Exception):
    pass


def pack_one_packet(packet: tuple) -> bytes:
    ee = get_enabled_encoders()
    if ee:
        e = get_encoder(ee[0])
        data, flags = e(packet)
        return pack_header(flags, 0, 0, len(data)) + data
    return strtobytes(packet)


def decode(data, protocol_flags: int):
    if isinstance(data, memoryview):
        data = data.tobytes()
    ptype = get_packet_encoding_type(protocol_flags)
    e = ENCODERS.get(ptype)
    if e:
        return e.decode(data)
    raise InvalidPacketEncodingException(f"{ptype!r} decoder is not available")


def main():  # pragma: no cover
    from xpra.util.str_fn import print_nested_dict
    from xpra.platform import program_context
    with program_context("Packet Encoding", "Packet Encoding Info"):
        init_all()
        print_nested_dict(get_packet_encoding_caps())


if __name__ == "__main__":  # pragma: no cover
    main()
